# mypy: ignore-errors
"""
==========================
Vivarium Testing Utilities
==========================

Utility functions and classes to make testing ``vivarium`` components easier.

"""

from itertools import product
from pathlib import Path
from typing import Any, Dict, List

import numpy as np
import pandas as pd

from vivarium import Component
from vivarium.framework import randomness
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.population import SimulantData
from vivarium.framework.randomness.index_map import IndexMap


class NonCRNTestPopulation(Component):
    CONFIGURATION_DEFAULTS = {
        "population": {
            "initialization_age_min": 0,
            "initialization_age_max": 100,
            "untracking_age": None,
        },
    }

    @property
    def columns_created(self) -> List[str]:
        return ["age", "sex", "location", "alive", "entrance_time", "exit_time"]

    def setup(self, builder: Builder) -> None:
        self.config = builder.configuration
        self.randomness = builder.randomness.get_stream(
            "population_age_fuzz", initializes_crn_attributes=True
        )

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        age_start = pop_data.user_data.get(
            "age_start", self.config.population.initialization_age_min
        )
        age_end = pop_data.user_data.get(
            "age_end", self.config.population.initialization_age_max
        )
        location = self.config.input_data.location

        population = _non_crn_build_population(
            pop_data.index,
            age_start,
            age_end,
            location,
            pop_data.creation_time,
            pop_data.creation_window,
            self.randomness,
        )
        self.population_view.update(population)

    def on_time_step(self, event: Event) -> None:
        population = self.population_view.get(event.index, query="alive == 'alive'")
        population["age"] += event.step_size / pd.Timedelta(days=365)
        self.population_view.update(population)


class TestPopulation(NonCRNTestPopulation):
    def setup(self, builder: Builder) -> None:
        super().setup(builder)
        self.age_randomness = builder.randomness.get_stream(
            "age_initialization", initializes_crn_attributes=True
        )
        self.register = builder.randomness.register_simulants

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        age_start = pop_data.user_data.get(
            "age_start", self.config.population.initialization_age_min
        )
        age_end = pop_data.user_data.get(
            "age_end", self.config.population.initialization_age_max
        )
        age_draw = self.age_randomness.get_draw(pop_data.index)
        if age_start == age_end:
            age = age_draw * (pop_data.creation_window / pd.Timedelta(days=365)) + age_start
        else:
            age = age_draw * (age_end - age_start) + age_start

        core_population = pd.DataFrame(
            {"entrance_time": pop_data.creation_time, "age": age.values}, index=pop_data.index
        )
        self.register(core_population)

        if "location" in self.config.input_data.keys():
            location = self.config.input_data.location
        else:
            location = self.randomness.choice(
                pop_data.index, ["USA", "Canada", "Mexico"], additional_key="location_choice"
            )
        population = _build_population(core_population, location, self.randomness)
        self.population_view.update(population)


def _build_population(core_population, location, randomness_stream):
    index = core_population.index

    population = pd.DataFrame(
        {
            "age": core_population["age"],
            "entrance_time": core_population["entrance_time"],
            "sex": randomness_stream.choice(
                index, ["Male", "Female"], additional_key="sex_choice"
            ),
            "alive": pd.Series("alive", index=index),
            "location": location,
            "exit_time": pd.NaT,
        },
        index=index,
    )
    return population


def _non_crn_build_population(
    index, age_start, age_end, location, creation_time, creation_window, randomness_stream
):
    if age_start == age_end:
        age = (
            randomness_stream.get_draw(index) * (creation_window / pd.Timedelta(days=365))
            + age_start
        )
    else:
        age = randomness_stream.get_draw(index) * (age_end - age_start) + age_start

    population = pd.DataFrame(
        {
            "age": age,
            "sex": randomness_stream.choice(
                index, ["Male", "Female"], additional_key="sex_choice"
            ),
            "alive": pd.Series("alive", index=index),
            "location": location,
            "entrance_time": creation_time,
            "exit_time": pd.NaT,
        },
        index=index,
    )
    return population


def build_table(
    value: Any,
    parameter_columns: Dict = {
        "age": (0, 125),
        "year": (1990, 2020),
    },
    key_columns: Dict = {"sex": ("Female", "Male")},
    value_columns: List = ["value"],
) -> pd.DataFrame:
    """

    Parameters
    ----------
    value
        Value(s) to put in the value columns of a lookup table.
    parameter_columns
        A dictionary where the keys are parameter (continuous) columns of a lookup table
        and the values are tuple of the range (inclusive) for that column.
    key_columns
        A dictionary where the keys are key (categorical) columns of a lookup table
        and the values are a tuple of the categories for that column
    value_columns
        A list of value columns that will appear in the returned lookup table

    Returns
    -------
        A pandas dataframe that has the cartesian product of the range of all parameter columns
        and the values of the key columns.
    """
    if not isinstance(value, list):
        value = [value] * len(value_columns)

    if len(value) != len(value_columns):
        raise ValueError("Number of values must match number of value columns")

    # Get product of parameter columns
    range_parameter_product = {
        key: list(range(value[0], value[1])) for key, value in parameter_columns.items()
    }
    # Build out dict of items we will need cartesian product of to make dataframe
    product_dict = dict(range_parameter_product)
    product_dict.update(key_columns)
    products = product(*product_dict.values())

    rows = []
    for item in products:
        # Note: item is going to be a tuple of the cartesian product of the key column values and parameter column
        # values and will be ordered in the order of the parameter then key dict keys
        r_values = []
        for val in value:
            if val is None:
                r_values.append(np.random.random())
            elif callable(val):
                r_values.append(val(item))
            else:
                r_values.append(val)

        # Get list of values for rows (index values)
        key_columns_index_values = list(item[len(parameter_columns) :])
        # Transform parameter column values
        parameter_columns_index_values = item[: len(parameter_columns)]
        # Create intervals for parameter columns. Example year, year+1 for year_start and year_end
        parameter_columns_index_values = [
            v for val in parameter_columns_index_values for v in (val, val + 1)
        ]
        rows.append(parameter_columns_index_values + key_columns_index_values + r_values)

    # Make list of parameter column names
    parameter_column_names = [
        col_name for col in parameter_columns for col_name in (f"{col}_start", f"{col}_end")
    ]

    return pd.DataFrame(
        rows,
        columns=parameter_column_names + list(key_columns.keys()) + value_columns,
    )


def make_dummy_column(name, initial_value):
    class DummyColumnMaker:
        @property
        def name(self):
            return "dummy_column_maker"

        def setup(self, builder):
            self.population_view = builder.population.get_view([name])
            builder.population.initializes_simulants(self.make_column, creates_columns=[name])

        def make_column(self, pop_data):
            self.population_view.update(
                pd.Series(initial_value, index=pop_data.index, name=name)
            )

        def __repr__(self):
            return f"dummy_column(name={name}, initial_value={initial_value})"

    return DummyColumnMaker()


def get_randomness(
    key="test",
    clock=lambda: pd.Timestamp(1990, 7, 2),
    seed=12345,
    initializes_crn_attributes=False,
):
    return randomness.RandomnessStream(
        key,
        clock,
        seed=seed,
        index_map=IndexMap(),
        initializes_crn_attributes=initializes_crn_attributes,
    )


def reset_mocks(mocks):
    for mock in mocks:
        mock.reset_mock()


def metadata(file_path, layer="override"):
    return {"layer": layer, "source": str(Path(file_path).resolve())}



import subprocess

import pandas as pd


def test_reproducibility(tmp_path, disease_model_spec):

    results_dir = tmp_path / "repro_check"
    results_dir.mkdir()

    cmd = f"simulate run {str(disease_model_spec)} -o {str(results_dir)}"

    subprocess.run(
        cmd,
        shell=True,
        check=True,
    )

    subprocess.run(
        cmd,
        shell=True,
        check=True,
    )

    files = [file for file in results_dir.rglob("**/*.parquet")]
    assert len(files) == 4
    for filename in ["dead", "ylls"]:
        df_paths = [file for file in files if file.stem == filename]
        df1 = pd.read_parquet(df_paths[0])
        df2 = pd.read_parquet(df_paths[1])

        assert df1.equals(df2)



from itertools import chain, combinations, product

import numpy as np
import pandas as pd
import pytest
from scipy.stats import chisquare

from vivarium.framework.randomness import RandomnessError
from vivarium.framework.randomness.index_map import IndexMap


def almost_powerset(iterable):
    """almost_powerset([1,2,3]) --> (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    return list(chain.from_iterable(combinations(s, r) for r in range(1, len(s) + 1)))


def generate_keys(number, types=("int", "float", "datetime"), seed=123456):
    rs = np.random.RandomState(seed=seed)

    keys = {}
    if "datetime" in types:
        year = rs.choice(np.arange(1980, 2018))
        day = rs.choice(pd.date_range(f"01/01/{year}", periods=365))
        start_time = rs.choice(pd.date_range(day, periods=86400, freq="s"))
        freq = rs.choice(["ms", "s", "min", "h"])
        keys["datetime"] = pd.date_range(start_time, periods=number, freq=freq)

    if "int" in types:
        kind = rs.choice(["random", "sequential"])
        if kind == "random":
            ints = np.unique(rs.randint(0, 1000 * number, size=100 * number))
            assert len(ints) > number
            rs.shuffle(ints)
            keys["int"] = ints[:number]
        else:
            start = rs.randint(0, 100 * number)
            keys["int"] = np.arange(start, start + number, dtype=int)

    if "float" in types:
        keys["float"] = rs.random_sample(size=number)

    return pd.DataFrame(keys, index=pd.RangeIndex(number))


rs = np.random.RandomState(seed=456789)
index_sizes = list(rs.randint(10_000, 250_000, size=1))
types = list(almost_powerset(["int", "float", "datetime"]))
seeds = list(rs.randint(10000000, size=1))


def id_fun(param):
    return f"Size:{param[0]}, Types:{param[1]}, Seed:{param[2]}"


@pytest.fixture(scope="module", params=list(product(index_sizes, types, seeds)), ids=id_fun)
def map_size_and_hashed_values(request):
    index_size, types_, seed = request.param
    keys = generate_keys(*request.param).set_index(types_).index
    m = IndexMap(key_columns=types_)
    return len(m), m._hash(keys)


def test_digit_scalar():
    m = IndexMap()
    k = 123456789
    for i in range(10):
        assert m._digit(k, i) == 10 - (i + 1)


def test_digit_series():
    m = IndexMap()
    k = pd.Series(123456789, index=range(10000))
    for i in range(10):
        assert len(m._digit(k, i).unique()) == 1
        assert m._digit(k, i)[0] == 10 - (i + 1)


def test_clip_to_seconds_scalar():
    m = IndexMap()
    k = pd.to_datetime("2010-01-25 06:25:31.123456789")
    assert m._clip_to_seconds(k.value) == int(str(k.value)[:10])


def test_clip_to_seconds_series():
    m = IndexMap()
    stamp = 1234567890
    k = (
        pd.date_range(pd.to_datetime(stamp, unit="s"), periods=10000, freq="ns")
        .to_series()
        .astype(np.int64)
    )
    assert len(m._clip_to_seconds(k).unique()) == 1
    assert m._clip_to_seconds(k).unique()[0] == stamp


def test_spread_scalar():
    m = IndexMap()
    assert m._spread(1234567890) == 4072825790


def test_spread_series():
    m = IndexMap()
    s = pd.Series(1234567890, index=range(10000))
    assert len(m._spread(s).unique()) == 1
    assert m._spread(s).unique()[0] == 4072825790


def test_shift_scalar():
    m = IndexMap()
    assert m._shift(1.1234567890) == 1234567890


def test_shift_series():
    m = IndexMap()
    s = pd.Series(1.1234567890, index=range(10000))
    assert len(m._shift(s).unique()) == 1
    assert m._shift(s).unique()[0] == 1234567890


def test_convert_to_ten_digit_int():
    m = IndexMap()
    v = 1234567890
    datetime_col = pd.date_range(
        pd.to_datetime(v, unit="s"), periods=10000, freq="ns"
    ).to_series()
    int_col = pd.Series(v, index=range(10000))
    float_col = pd.Series(1.1234567890, index=range(10000))
    bad_col = pd.Series("a", index=range(10000))

    assert m._convert_to_ten_digit_int(datetime_col).unique()[0] == v
    assert m._convert_to_ten_digit_int(int_col).unique()[0] == 4072825790
    assert m._convert_to_ten_digit_int(float_col).unique()[0] == v
    with pytest.raises(RandomnessError):
        m._convert_to_ten_digit_int(bad_col)


@pytest.mark.skip("This fails because the hash needs work")
def test_hash_collisions(map_size_and_hashed_values):
    n, h = map_size_and_hashed_values
    k = len(h)

    expected_empty_bins = n * (1 - 1 / n) ** k  # Potential source of roundoff issues.
    expected_full_bins = n - expected_empty_bins
    expected_collisions = k - expected_full_bins

    uniques = h.drop_duplicates()
    actual_collisions = len(h) - len(uniques)

    assert actual_collisions / expected_collisions < 1.5, "Too many collisions"


@pytest.mark.skip("This fails because the hash needs work")
def test_hash_uniformity(map_size_and_hashed_values):
    n, h = map_size_and_hashed_values

    k = len(h)
    num_bins = k // 5  # Want about 5 items per bin for chi-squared
    bins = np.linspace(0, n + 1, num_bins)

    binned_data = pd.cut(h, bins)
    distribution = pd.value_counts(binned_data).sort_index()
    c, p = chisquare(distribution)

    assert p > 0.05, "Data not uniform"


@pytest.fixture(scope="function")
def index_map(mocker):
    mock_index_map = IndexMap

    def hash_mock(k, salt):
        seed = 123456
        salt = IndexMap()._convert_to_ten_digit_int(pd.Series(salt, index=k))
        rs = np.random.RandomState(seed=seed + salt)
        return pd.Series(rs.randint(0, len(k) * 10, size=len(k)), index=k)

    mocker.patch.object(mock_index_map, "_hash", side_effect=hash_mock)

    return mock_index_map


def test_update_empty_bad_keys(index_map):
    keys = pd.DataFrame({"A": ["a"] * 10}, index=range(10))
    m = index_map(key_columns=list(keys.columns))
    with pytest.raises(RandomnessError):
        m.update(keys, pd.to_datetime("2023-01-01"))


def test_update_nonempty_bad_keys(index_map):
    keys = generate_keys(1000)
    m = index_map(key_columns=list(keys.columns))
    m.update(keys, pd.to_datetime("2023-01-01"))
    with pytest.raises(RandomnessError):
        m.update(keys, pd.to_datetime("2023-01-01"))


def test_update_empty_good_keys(index_map):
    keys = generate_keys(1000)
    m = index_map(key_columns=list(keys.columns))
    m.update(keys, pd.to_datetime("2023-01-01"))
    key_index = keys.set_index(list(keys.columns)).index
    assert len(m._map) == len(keys), "All keys not in mapping"
    assert (
        m._map.index.droplevel(m.SIM_INDEX_COLUMN).difference(key_index).empty
    ), "Extra keys in mapping"
    assert len(m._map.unique()) == len(keys), "Duplicate values in mapping"


def test_update_nonempty_good_keys(index_map):
    keys = generate_keys(2000)
    m = index_map(key_columns=list(keys.columns))
    keys1, keys2 = keys[:1000], keys[1000:]

    m.update(keys1, pd.to_datetime("2023-01-01"))
    m.update(keys2, pd.to_datetime("2023-01-01"))

    key_index = keys.set_index(list(keys.columns)).index
    assert len(m._map) == len(keys), "All keys not in mapping"
    assert (
        m._map.index.droplevel(m.SIM_INDEX_COLUMN).difference(key_index).empty
    ), "Extra keys in mapping"
    assert len(m._map.unique()) == len(keys), "Duplicate values in mapping"



import numpy as np
import pandas as pd
import pytest

from vivarium.framework.utilities import from_yearly
from vivarium.framework.values import (
    ValuesManager,
    list_combiner,
    rescale_post_processor,
    union_post_processor,
)


@pytest.fixture
def static_step():
    return lambda idx: pd.Series(pd.Timedelta(days=6), index=idx)


@pytest.fixture
def variable_step():
    return lambda idx: pd.Series(
        [pd.Timedelta(days=3) if i % 2 == 0 else pd.Timedelta(days=5) for i in idx], index=idx
    )


@pytest.fixture
def manager(mocker):
    manager = ValuesManager()
    builder = mocker.MagicMock()
    manager.setup(builder)
    return manager


@pytest.fixture
def manager_with_step_size(mocker, request):
    manager = ValuesManager()
    builder = mocker.MagicMock()
    builder.time.step_size = lambda: lambda: pd.Timedelta(days=6)
    builder.time.simulant_step_sizes = lambda: request.getfixturevalue(request.param)
    manager.setup(builder)
    return manager


def test_replace_combiner(manager):
    value = manager.register_value_producer("test", source=lambda: 1)

    assert value() == 1

    manager.register_value_modifier("test", modifier=lambda v: 42)
    assert value() == 42

    manager.register_value_modifier("test", lambda v: 84)
    assert value() == 84


def test_joint_value(manager):
    # This is the normal configuration for PAF and disability weight type values
    index = pd.Index(range(10))

    value = manager.register_value_producer(
        "test",
        source=lambda idx: [pd.Series(0.0, index=idx)],
        preferred_combiner=list_combiner,
        preferred_post_processor=union_post_processor,
    )
    assert np.all(value(index) == 0)

    manager.register_value_modifier("test", modifier=lambda idx: pd.Series(0.5, index=idx))
    assert np.all(value(index) == 0.5)

    manager.register_value_modifier("test", modifier=lambda idx: pd.Series(0.5, index=idx))
    assert np.all(value(index) == 0.75)


def test_contains(manager):
    value = "test_value"
    rate = "test_rate"

    assert value not in manager
    assert rate not in manager

    manager.register_value_producer("test_value", source=lambda: 1)
    assert value in manager
    assert rate not in manager


def test_returned_series_name(manager):
    value = manager.register_value_producer(
        "test",
        source=lambda idx: pd.Series(0.0, index=idx),
    )
    assert value(pd.Index(range(10))).name == "test"


@pytest.mark.parametrize("manager_with_step_size", ["static_step"], indirect=True)
def test_rescale_post_processor_static(manager_with_step_size):
    index = pd.Index(range(10))

    pipeline = manager_with_step_size.register_value_producer(
        "test",
        source=lambda idx: pd.Series(0.75, index=idx),
        preferred_post_processor=rescale_post_processor,
    )
    assert np.all(pipeline(index) == from_yearly(0.75, pd.Timedelta(days=6)))


@pytest.mark.parametrize("manager_with_step_size", ["variable_step"], indirect=True)
def test_rescale_post_processor_variable(manager_with_step_size):
    index = pd.Index(range(10))

    pipeline = manager_with_step_size.register_value_producer(
        "test",
        source=lambda idx: pd.Series(0.5, index=idx),
        preferred_post_processor=rescale_post_processor,
    )
    value = pipeline(index)
    evens = value.iloc[lambda x: x.index % 2 == 0]
    odds = value.iloc[lambda x: x.index % 2 == 1]
    assert np.all(evens == from_yearly(0.5, pd.Timedelta(days=3)))
    assert np.all(odds == from_yearly(0.5, pd.Timedelta(days=5)))



import random
from pathlib import Path

import pandas as pd
import pytest

from vivarium.framework.artifact.manager import (
    ArtifactManager,
    _config_filter,
    _subset_columns,
    _subset_rows,
    parse_artifact_path_config,
    validate_filter_term,
)
from vivarium.testing_utilities import build_table, metadata


@pytest.fixture()
def artifact_mock(mocker):
    mock = mocker.patch("vivarium.framework.artifact.manager.Artifact")

    def mock_load(key):
        if key == "string_data.key":
            return "string_data"
        elif key == "df_data.key":
            return pd.DataFrame()
        else:
            return None

    mock.load.side_effect = mock_load

    return mock


def test_subset_rows_extra_filters():
    data = build_table(1)
    with pytest.raises(ValueError):
        _subset_rows(data, missing_thing=12)


def test_subset_rows():
    values = [
        lambda *args, **kwargs: random.choice(["red", "blue"]),
        lambda *args, **kwargs: random.choice([1, 2, 3]),
    ]
    data = build_table(
        value=values,
        value_columns=["color", "number"],
    )
    filtered_data = _subset_rows(data, color="red", number=3)
    assert filtered_data.equals(data[(data.color == "red") & (data.number == 3)])

    filtered_data = _subset_rows(data, color="red", number=[2, 3])
    assert filtered_data.equals(
        data[(data.color == "red") & ((data.number == 2) | (data.number == 3))]
    )


def test_subset_columns():
    values = [0, "red", 100]
    data = build_table(
        values,
        parameter_columns={
            "age": (0, 99),
            "year": (1990, 2010),
        },
        value_columns=["draw", "color", "value"],
    )

    filtered_data = _subset_columns(data)
    assert filtered_data.equals(
        data[["age_start", "age_end", "year_start", "year_end", "sex", "color", "value"]]
    )

    filtered_data = _subset_columns(data, color="red")
    assert filtered_data.equals(
        data[["age_start", "age_end", "year_start", "year_end", "sex", "value"]]
    )


def test_parse_artifact_path_config(base_config, test_data_dir):
    artifact_path = test_data_dir / "artifact.hdf"
    base_config.update(
        {"input_data": {"artifact_path": str(artifact_path)}}, **metadata(str(Path("/")))
    )

    assert parse_artifact_path_config(base_config) == str(artifact_path)


def test_parse_artifact_path_relative_no_source(base_config):
    artifact_path = "./artifact.hdf"
    base_config.update({"input_data": {"artifact_path": str(artifact_path)}})

    with pytest.raises(ValueError):
        parse_artifact_path_config(base_config)


def test_parse_artifact_path_relative(base_config, test_data_dir):
    base_config.update(
        {"input_data": {"artifact_path": "../../test_data/artifact.hdf"}},
        **metadata(__file__)
    )
    assert parse_artifact_path_config(base_config) == str(test_data_dir / "artifact.hdf")


def test_parse_artifact_path_config_fail(base_config):
    artifact_path = Path(__file__).parent / "not_an_artifact.hdf"
    base_config.update(
        {"input_data": {"artifact_path": str(artifact_path)}}, **metadata(str(Path("/")))
    )

    with pytest.raises(FileNotFoundError):
        parse_artifact_path_config(base_config)


def test_parse_artifact_path_config_fail_relative(base_config):
    base_config.update(
        {"input_data": {"artifact_path": "./not_an_artifact.hdf"}}, **metadata(__file__)
    )

    with pytest.raises(FileNotFoundError):
        parse_artifact_path_config(base_config)


def test_load_with_string_data(artifact_mock):
    am = ArtifactManager()
    am.artifact = artifact_mock
    am.config_filter_term = None
    assert am.load("string_data.key") == "string_data"


def test_load_with_no_data(artifact_mock):
    am = ArtifactManager()
    am.artifact = artifact_mock
    assert am.load("no_data.key") is None


def test_load_with_df_data(artifact_mock):
    am = ArtifactManager()
    am.artifact = artifact_mock
    am.config_filter_term = None
    assert isinstance(am.load("df_data.key"), pd.DataFrame)


def test_config_filter():
    df = pd.DataFrame({"year": range(1990, 2000, 1), "color": ["red", "yellow"] * 5})
    filtered = _config_filter(df, "year in [1992, 1995]")

    assert set(filtered.year) == {1992, 1995}


def test_config_filter_on_nonexistent_column():
    df = pd.DataFrame({"year": range(1990, 2000, 1), "color": ["red", "yellow"] * 5})
    filtered = _config_filter(df, "fake_col in [1992, 1995]")

    assert df.equals(filtered)


def test_validate_filter_term():
    multiple_filter_terms = "draw in [0, 1] and year > 1990"

    with pytest.raises(NotImplementedError):
        validate_filter_term(multiple_filter_terms)



# mypy: ignore-errors
"""
===================
The Population View
===================

The :class:`PopulationView` is a user-facing abstraction that manages read and write access
to the underlying simulation :term:`State Table`. It has two primary responsibilities:

    1. To provide user access to subsets of the simulation state table
       when it is safe to do so.
    2. To allow the user to update the simulation state in a controlled way.

"""

from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import pandas as pd

from vivarium.framework.population.exceptions import PopulationError

if TYPE_CHECKING:
    # Cyclic import
    from vivarium.framework.population.manager import PopulationManager


class PopulationView:
    """A read/write manager for the simulation state table.

    It can be used to both read and update the state of the population. A
    PopulationView can only read and write columns for which it is configured.
    Attempts to update non-existent columns are ignored except during
    simulant creation when new columns are allowed to be created.

    Notes
    -----
    By default, this view will filter out ``untracked`` simulants unless
    the ``tracked`` column is specified in the initialization arguments.

    """

    def __init__(
        self,
        manager: "PopulationManager",
        view_id: int,
        columns: Union[List[str], Tuple[str]] = (),
        query: Optional[str] = None,
    ):
        """

        Parameters
        ----------
        manager
            The population manager for the simulation.
        view_id
            The unique identifier for this view.
        columns
            The set of columns this view should have access too.  If empty, this
            view will have access to the entire state table.
        query
            A :mod:`pandas`-style filter that will be applied any time this
            view is read from.
        """
        self._manager = manager
        self._id = view_id
        self._columns = list(columns)
        self._query = query

    @property
    def name(self):
        return f"population_view_{self._id}"

    @property
    def columns(self) -> List[str]:
        """The columns that the view can read and update.

        If the view was created with ``None`` as the columns argument, then
        the view will have access to the full table by default. That case
        should be only be used in situations where the full state table is
        actually needed, like for some metrics collection applications.
        """
        if not self._columns:
            return list(self._manager.get_population(True).columns)
        return list(self._columns)

    @property
    def query(self) -> Optional[str]:
        """A :mod:`pandas` style query to filter the rows of this view.

        This query will be applied any time the view is read. This query may
        reference columns not in the view's columns.
        """
        return self._query

    def subview(self, columns: Union[List[str], Tuple[str]]) -> "PopulationView":
        """Retrieves a new view with a subset of this view's columns.

        Parameters
        ----------
        columns
            The set of columns to provide access to in the subview. Must be
            a proper subset of this view's columns.

        Returns
        -------
            A new view with access to the requested columns.

        Raises
        ------
        PopulationError
            If the requested columns are not a proper subset of this view's
            columns or no columns are requested.

        Notes
        -----
        Subviews are useful during population initialization. The original
        view may contain both columns that a component needs to create and
        update as well as columns that the component needs to read.  By
        requesting a subview, a component can read the sections it needs
        without running the risk of trying to access uncreated columns
        because the component itself has not created them.
        """

        if not columns or set(columns) - set(self.columns):
            raise PopulationError(
                f"Invalid subview requested.  Requested columns must be a non-empty "
                f"subset of this view's columns.  Requested columns: {columns}, "
                f"Available columns: {self.columns}"
            )
        # Skip constraints for requesting subviews.
        return self._manager._get_view(columns, self.query)

    def get(self, index: pd.Index, query: str = "") -> pd.DataFrame:
        """Select the rows represented by the given index from this view.

        For the rows in ``index`` get the columns from the simulation's
        state table to which this view has access. The resulting rows may be
        further filtered by the view's query and only return a subset
        of the population represented by the index.

        Parameters
        ----------
        index
            Index of the population to get.
        query
            Additional conditions used to filter the index. These conditions
            will be unioned with the default query of this view.  The query
            provided may use columns that this view does not have access to.

        Returns
        -------
            A table with the subset of the population requested.

        Raises
        ------
        PopulationError
            If this view has access to columns that have not yet been created
            and this method is called.  If you see this error, you should
            request a subview with the columns you need read access to.

        See Also
        --------
        :meth:`subview <PopulationView.subview>`
        """
        pop = self._manager.get_population(True).loc[index]

        if not index.empty:
            if self._query:
                pop = pop.query(self._query)
            if query:
                pop = pop.query(query)

        non_existent_columns = set(self.columns) - set(pop.columns)
        if non_existent_columns:
            raise PopulationError(
                f"Requested column(s) {non_existent_columns} not in population table. "
                "This is likely due to a failure to require some columns, randomness "
                "streams, or pipelines when registering a simulant initializer, a value "
                "producer, or a value modifier. NOTE: It is possible for a run to "
                "succeed even if resource requirements were not properly specified in "
                "the simulant initializers or pipeline creation/modification calls. This "
                "success depends on component initialization order which may change in "
                "different run settings."
            )
        return pop.loc[:, self.columns]

    def update(self, population_update: Union[pd.DataFrame, pd.Series]) -> None:
        """Updates the state table with the provided data.

        Parameters
        ----------
        population_update
            The data which should be copied into the simulation's state. If
            the update is a :class:`pandas.DataFrame`, it can contain a subset
            of the view's columns but no extra columns. If ``pop`` is a
            :class:`pandas.Series` it must have a name that matches one of
            this view's columns unless the view only has one column in which
            case the Series will be assumed to refer to that regardless of its
            name.

        Raises
        ------
        PopulationError
            If the provided data name or columns do not match columns that
            this view manages or if the view is being updated with a data
            type inconsistent with the original population data.
        """
        state_table = self._manager.get_population(True)
        population_update = self._format_update_and_check_preconditions(
            population_update,
            state_table,
            self.columns,
            self._manager.creating_initial_population,
            self._manager.adding_simulants,
        )
        if self._manager.creating_initial_population:
            new_columns = list(set(population_update).difference(state_table))
            self._manager._population[new_columns] = population_update[new_columns]
        elif not population_update.empty:
            update_columns = list(set(population_update).intersection(state_table))
            for column in update_columns:
                column_update = self._update_column_and_ensure_dtype(
                    population_update[column],
                    state_table[column],
                    self._manager.adding_simulants,
                )
                self._manager._population[column] = column_update

    def __repr__(self):
        return (
            f"PopulationView(_id={self._id}, _columns={self.columns}, _query={self._query})"
        )

    ##################
    # Helper methods #
    ##################

    @staticmethod
    def _format_update_and_check_preconditions(
        population_update: Union[pd.Series, pd.DataFrame],
        state_table: pd.DataFrame,
        view_columns: List[str],
        creating_initial_population: bool,
        adding_simulants: bool,
    ) -> pd.DataFrame:
        """Standardizes the population update format and checks preconditions.

        Managing how values get written to the underlying population state table is critical
        to rule out several categories of error in client simulation code. The state table
        is modified at three different times. In the first, the initial population table
        is being created and new columns are being added to the state table with their
        initial values. In the second, the population manager has added new rows with
        appropriate null values to the state table in response to population creation
        dictated by client code, and population updates are being provided to fill in
        initial values for those new rows. In the final case, state table values for
        existing simulants are being overridden as part of a time step.

        All of these modification scenarios require that certain preconditions are met.
        For all scenarios, we require

            1. The update is a DataFrame or a Series.
            2. If it is a series, it is nameless and this view manages a single column
               or it is named and it's name matches a column in this PopulationView.
            3. The update matches at least one column in this PopulationView.
            4. The update columns are a subset of the columns managed by this
               PopulationView.
            5. The update index is a subset of the existing state table index.
               PopulationViews don't make rows, they just fill them in.

        For initial population creation additional preconditions are documented in
        :meth:`PopulationView._ensure_coherent_initialization`. Outside population
        initialization, we require that all columns in the update to be present in
        the existing state table. When new simulants are added in the middle of the
        simulation, we require that only one component provide updates to a column.

        Parameters
        ----------
        population_update
            The update to the simulation state table.
        state_table
            The existing simulation state table.
        view_columns
            The columns managed by this PopulationView.
        creating_initial_population
            Whether the initial population is being created.
        adding_simulants
            Whether new simulants are currently being initialized.

        Returns
        -------
            The input data formatted as a DataFrame.

        Raises
        ------
        TypeError
            If the population update is not a :class:`pandas.Series` or a
            :class:`pandas.DataFrame`.
        PopulationError
            If the update violates any preconditions relevant to the context in which
            the update is provided (initial population creation, population creation on
            time steps, or population state changes on time steps).

        """
        assert not creating_initial_population or adding_simulants

        population_update = PopulationView._coerce_to_dataframe(
            population_update,
            view_columns,
        )

        unknown_simulants = len(population_update.index.difference(state_table.index))
        if unknown_simulants:
            raise PopulationError(
                "Population updates must have an index that is a subset of the current "
                f"population state table. {unknown_simulants} simulants were provided "
                f"in an update with no matching index in the existing table."
            )

        if creating_initial_population:
            PopulationView._ensure_coherent_initialization(population_update, state_table)
        else:
            new_columns = list(set(population_update).difference(state_table))
            if new_columns:
                raise PopulationError(
                    f"Attempting to add new columns {new_columns} to the state table "
                    f"outside the initial population creation phase."
                )

            if adding_simulants:
                state_table_new_simulants = state_table.loc[population_update.index, :]
                conflicting_columns = [
                    column
                    for column in population_update
                    if state_table_new_simulants[column].notnull().any()
                    and not population_update[column].equals(
                        state_table_new_simulants[column]
                    )
                ]
                if conflicting_columns:
                    raise PopulationError(
                        "Two components are providing conflicting initialization data "
                        f"for the state table columns: {conflicting_columns}."
                    )

        return population_update

    @staticmethod
    def _coerce_to_dataframe(
        population_update: Union[pd.Series, pd.DataFrame],
        view_columns: List[str],
    ) -> pd.DataFrame:
        """Coerce all population updates to a :class:`pandas.DataFrame` format.

        Parameters
        ----------
        population_update
            The update to the simulation state table.

        Returns
        -------
            The input data formatted as a DataFrame.

        Raises
        ------
        TypeError
            If the population update is not a :class:`pandas.Series` or a
            :class:`pandas.DataFrame`.
        PopulationError
            If the input data is a :class:`pandas.Series` and this :class:`PopulationView`
            manages multiple columns or if the population update contains columns not
            managed by this view.
        """
        if not isinstance(population_update, (pd.Series, pd.DataFrame)):
            raise TypeError(
                "The population update must be a pandas Series or DataFrame. "
                f"A {type(population_update)} was provided."
            )

        if isinstance(population_update, pd.Series):
            if population_update.name is None:
                if len(view_columns) == 1:
                    population_update.name = view_columns[0]
                else:
                    raise PopulationError(
                        "Cannot update with an unnamed pandas series unless there "
                        "is only a single column in the view."
                    )

            population_update = pd.DataFrame(population_update)

        if not set(population_update.columns).issubset(view_columns):
            raise PopulationError(
                f"Cannot update with a DataFrame or Series that contains columns "
                f"the view does not. Dataframe contains the following extra columns: "
                f"{set(population_update.columns).difference(view_columns)}."
            )

        update_columns = list(population_update)
        if not update_columns:
            raise PopulationError(
                "The update method of population view is being called "
                "on a DataFrame with no columns."
            )

        return population_update

    @staticmethod
    def _ensure_coherent_initialization(
        population_update: pd.DataFrame, state_table: pd.DataFrame
    ) -> None:
        """Ensure that overlapping population updates have the same information.

        During population initialization, each state table column should be updated by
        exactly one component and each component with an initializer should create at
        least one column. Sometimes components are a little sloppy and provide
        duplicate column information, which we should continue to allow. We want to ensure
        that a column is only getting one set of unique values though.

        Parameters
        ----------
        population_update
            The update to the simulation state table.
        state_table
            The existing simulation state table. When this method is called, the table
            should be in a partially complete state. That is the provided population
            update should carry some new attributes we need to assign.

        Raises
        -----
        PopulationError
            If the population update contains no new information or if it contains
            information in conflict with the existing state table.
        """
        missing_pops = len(state_table.index.difference(population_update.index))
        if missing_pops:
            raise PopulationError(
                f"Components should initialize the same population at the simulation start. "
                f"A component is missing updates for {missing_pops} simulants."
            )
        new_columns = set(population_update).difference(state_table)
        overlapping_columns = set(population_update).intersection(state_table)
        if not new_columns:
            raise PopulationError(
                f"A component is providing a population update for {list(population_update)} "
                "but all provided columns are initialized by other components."
            )
        for column in overlapping_columns:
            if not population_update[column].equals(state_table[column]):
                raise PopulationError(
                    "Two components are providing conflicting initialization data for the "
                    f"{column} state table column."
                )

    @staticmethod
    def _update_column_and_ensure_dtype(
        update: pd.Series,
        existing: pd.Series,
        adding_simulants: bool,
    ) -> pd.Series:
        """Build the updated state table column with an appropriate dtype.

        Parameters
        ----------
        update
            The new column values for a subset of the existing index.
        existing
            The existing column values for all simulants in the state table.
        adding_simulants
            Whether new simulants are currently being initialized.

        Returns
        -------
            The column with the provided update applied
        """
        # FIXME: This code does not work as described. I'm leaving it here because writing
        #  real dtype checking code is a pain and we never seem to hit the actual edge cases.
        #  I've also seen this error, though I don't have a reproducible and useful example.
        #  I'm reasonably sure what's really being accounted for here is non-nullable columns
        #  that temporarily have null values introduced in the space between rows being
        #  added to the state table and initializers filling them with their first values.
        #  That means the space of dtype casting issues is actually quite small. What should
        #  actually happen in the long term is to separate the population creation entirely
        #  from the mutation of existing state. I.e. there's not an actual reason we need
        #  to do all these sequential operations on a single underlying dataframe during
        #  the creation of new simulants besides the fact that it's the existing
        #  implementation.
        update_values = update.values.copy()
        new_state_table_values = existing.values.copy()
        update_index_positional = existing.index.get_indexer(update.index)

        # Assumes the update index labels can be interpreted as an array position.
        new_state_table_values[update_index_positional] = update_values

        unmatched_dtypes = new_state_table_values.dtype != update_values.dtype
        if unmatched_dtypes and not adding_simulants:
            # This happens when the population is being grown because extending
            # the index forces columns that don't have a natural null type
            # to become 'object'
            raise PopulationError(
                "A component is corrupting the population table by modifying the dtype of "
                f"the {update.name} column from {existing.dtype} to {update.dtype}."
            )
        new_state_table_values = new_state_table_values.astype(update_values.dtype)
        return pd.Series(new_state_table_values, index=existing.index, name=existing.name)



from datetime import datetime, timedelta
from numbers import Number
from typing import Union

import numpy as np
import numpy.typing as npt
import pandas as pd

NumericArray = npt.NDArray[np.number[npt.NBitBase]]

ScalarValue = Union[Number, timedelta, datetime]
LookupTableData = Union[ScalarValue, pd.DataFrame, list[ScalarValue], tuple[ScalarValue]]
# TODO: For some of the uses of NumberLike, we probably want a TypeVar here instead.
NumberLike = Union[
    NumericArray,
    # TODO: Parameterizing pandas objects fails below python 3.12
    pd.Series,  # type: ignore [type-arg]
    pd.DataFrame,
    float,
    int,
]
# todo need to use TypeVars here
Time = Union[pd.Timestamp, datetime]
Timedelta = Union[pd.Timedelta, timedelta]
ClockTime = Union[Time, Number]
ClockStepSize = Union[Timedelta, Number]



import itertools
import math
import re
from datetime import timedelta

import numpy as np
import pandas as pd
import pytest
from layered_config_tree.main import LayeredConfigTree
from loguru import logger
from pandas.core.groupby.generic import DataFrameGroupBy

from tests.framework.results.helpers import (
    BASE_POPULATION,
    FAMILIARS,
    HOUSE_CATEGORIES,
    NAME,
    NAME_COLUMNS,
    sorting_hat_serial,
    sorting_hat_vectorized,
    verify_stratification_added,
)
from vivarium.framework.event import Event
from vivarium.framework.results import VALUE_COLUMN
from vivarium.framework.results.context import ResultsContext
from vivarium.framework.results.observation import AddingObservation, ConcatenatingObservation


def _aggregate_state_person_time(x: pd.DataFrame) -> float:
    """Helper aggregator function for observation testing"""
    return len(x) * (28 / 365.25)


@pytest.fixture
def mocked_event(mocker) -> Event:
    event: Event = mocker.Mock(spec=Event)
    return event


@pytest.mark.parametrize(
    "mapper, is_vectorized",
    [
        (sorting_hat_vectorized, True),
        (sorting_hat_serial, False),
    ],
    ids=["vectorized_mapper", "non-vectorized_mapper"],
)
def test_add_stratification_mappers(mapper, is_vectorized, mocker):
    ctx = ResultsContext()
    mocker.patch.object(ctx, "excluded_categories", {})
    assert not verify_stratification_added(
        ctx.stratifications, NAME, NAME_COLUMNS, HOUSE_CATEGORIES, [], mapper, is_vectorized
    )
    ctx.add_stratification(
        name=NAME,
        sources=NAME_COLUMNS,
        categories=HOUSE_CATEGORIES,
        excluded_categories=None,
        mapper=mapper,
        is_vectorized=is_vectorized,
    )
    assert verify_stratification_added(
        ctx.stratifications, NAME, NAME_COLUMNS, HOUSE_CATEGORIES, [], mapper, is_vectorized
    )


@pytest.mark.parametrize(
    "excluded_categories",
    [
        [],
        HOUSE_CATEGORIES[:1],
        HOUSE_CATEGORIES[:2],
        HOUSE_CATEGORIES[:3],
    ],
    ids=[
        "no_excluded_categories",
        "one_excluded_category",
        "two_excluded_categories",
        "all_but_one_excluded_categories",
    ],
)
def test_add_stratification_excluded_categories(excluded_categories, mocker):
    ctx = ResultsContext()
    builder = mocker.Mock()
    builder.configuration.stratification = LayeredConfigTree(
        {"default": [], "excluded_categories": {NAME: excluded_categories}}
    )
    builder.logging.get_logger.return_value = logger
    ctx.setup(builder)
    assert not verify_stratification_added(
        ctx.stratifications,
        NAME,
        NAME_COLUMNS,
        HOUSE_CATEGORIES,
        [],
        sorting_hat_vectorized,
        True,
    )
    ctx.add_stratification(
        name=NAME,
        sources=NAME_COLUMNS,
        categories=HOUSE_CATEGORIES,
        excluded_categories=excluded_categories,
        mapper=sorting_hat_vectorized,
        is_vectorized=True,
    )
    assert verify_stratification_added(
        ctx.stratifications,
        NAME,
        NAME_COLUMNS,
        HOUSE_CATEGORIES,
        excluded_categories,
        sorting_hat_vectorized,
        True,
    )


@pytest.mark.parametrize(
    "name, categories, excluded_categories, msg_match",
    [
        (
            "duplicate_name",
            HOUSE_CATEGORIES,
            [],
            "Stratification name 'duplicate_name' is already used: ",
        ),
        (
            NAME,
            HOUSE_CATEGORIES + ["slytherin"],
            [],
            f"Found duplicate categories in stratification '{NAME}': ['slytherin']",
        ),
        (
            NAME,
            HOUSE_CATEGORIES + ["gryffindor", "slytherin"],
            [],
            f"Found duplicate categories in stratification '{NAME}': ['gryffindor', 'slytherin']",
        ),
        (
            NAME,
            HOUSE_CATEGORIES,
            ["gryfflepuff"],
            "Excluded categories {'gryfflepuff'} not found in categories",
        ),
    ],
    ids=[
        "duplicate_name",
        "duplicate_category",
        "duplicate_categories",
        "unknown_excluded_category",
    ],
)
def test_add_stratification_raises(name, categories, excluded_categories, msg_match, mocker):
    ctx = ResultsContext()
    mocker.patch.object(ctx, "excluded_categories", {name: excluded_categories})
    # Register a stratification to test against duplicate stratifications
    ctx.add_stratification(
        name="duplicate_name",
        sources=["foo"],
        categories=["bar"],
        excluded_categories=None,
        mapper=sorting_hat_serial,
        is_vectorized=False,
    )
    with pytest.raises(ValueError, match=re.escape(msg_match)):
        ctx.add_stratification(
            name=name,
            sources=NAME_COLUMNS,
            categories=categories,
            excluded_categories=excluded_categories,
            mapper=sorting_hat_vectorized,
            is_vectorized=True,
        )


@pytest.mark.parametrize(
    "kwargs",
    [
        {
            "name": "living_person_time",
            "pop_filter": 'alive == "alive" and undead == False',
            "when": "collect_metrics",
        },
        {
            "name": "undead_person_time",
            "pop_filter": "undead == True",
            "when": "time_step__prepare",
        },
    ],
    ids=["valid_on_collect_metrics", "valid_on_time_step__prepare"],
)
def test_register_observation(kwargs):
    ctx = ResultsContext()
    assert len(ctx.observations) == 0
    kwargs["results_formatter"] = lambda: None
    kwargs["stratifications"] = tuple()
    kwargs["aggregator_sources"] = []
    kwargs["aggregator"] = len
    ctx.register_observation(
        observation_type=AddingObservation,
        **kwargs,
    )
    assert len(ctx.observations) == 1


def test_register_observation_duplicate_name_raises():
    ctx = ResultsContext()
    ctx.register_observation(
        observation_type=AddingObservation,
        name="some-observation-name",
        pop_filter="some-pop-filter",
        when="some-when",
        results_formatter=lambda df: df,
        stratifications=(),
        aggregator_sources=[],
        aggregator=len,
    )
    with pytest.raises(
        ValueError, match="Observation name 'some-observation-name' is already used: "
    ):
        # register a different observation but w/ the same name
        ctx.register_observation(
            observation_type=ConcatenatingObservation,
            name="some-observation-name",
            pop_filter="some-other-pop-filter",
            when="some-other-when",
        )


@pytest.mark.parametrize(
    "pop_filter, aggregator_sources, aggregator, stratifications",
    [
        ("tracked==True", None, len, ["house", "familiar"]),
        (
            "tracked==True",
            ["power_level"],
            sum,
            ["house", "familiar"],
        ),
        (
            "tracked==True",
            [],
            _aggregate_state_person_time,
            ["house", "familiar"],
        ),
        ("tracked==True", None, len, ["house"]),
        ("tracked==True", ["power_level"], sum, ["house"]),
        (
            "tracked==True",
            [],
            _aggregate_state_person_time,
            ["house"],
        ),
    ],
    ids=[
        "len_aggregator_two_stratifications",
        "sum_aggregator_two_stratifications",
        "custom_aggregator_two_stratifications",
        "len_aggregator_one_stratification",
        "sum_aggregator_one_stratification",
        "custom_aggregator_one_stratification",
    ],
)
def test_adding_observation_gather_results(
    pop_filter, aggregator_sources, aggregator, stratifications, mocked_event
):
    """Test cases where every stratification is in gather_results. Checks for
    existence and correctness of results"""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()
    # Mock out some extra columns that would be produced by the manager's _prepare_population() method
    population["current_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12)
    population["event_step_size"] = timedelta(days=28)
    population["event_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12) + timedelta(
        days=28
    )
    lifecycle_phase = "collect_metrics"

    # Set up stratifications
    if "house" in stratifications:
        ctx.add_stratification(
            name="house",
            sources=["house"],
            categories=HOUSE_CATEGORIES,
            excluded_categories=None,
            mapper=None,
            is_vectorized=True,
        )
    if "familiar" in stratifications:
        ctx.add_stratification(
            name="familiar",
            sources=["familiar"],
            categories=FAMILIARS,
            excluded_categories=None,
            mapper=None,
            is_vectorized=True,
        )
    ctx.register_observation(
        observation_type=AddingObservation,
        name="foo",
        pop_filter=pop_filter,
        aggregator_sources=aggregator_sources,
        aggregator=aggregator,
        stratifications=tuple(stratifications),
        when=lifecycle_phase,
        results_formatter=lambda: None,
    )

    filtered_pop = population.query(pop_filter)
    groups = filtered_pop.groupby(stratifications)
    if aggregator == sum:
        power_level_sums = groups[aggregator_sources].sum().squeeze()
        assert len(power_level_sums.unique()) == 1
        expected_result = power_level_sums.iat[0]
    else:
        group_sizes = groups.size()
        assert len(group_sizes.unique()) == 1
        num_stratifications = group_sizes.iat[0]
        expected_result = (
            num_stratifications if aggregator == len else num_stratifications * 28 / 365.25
        )

    i = 0
    for result, _measure, _updater in ctx.gather_results(
        population, lifecycle_phase, mocked_event
    ):
        assert all(
            math.isclose(actual_result, expected_result, rel_tol=0.0001)
            for actual_result in result.values
        )
        i += 1
    assert i == 1


def test_concatenating_observation_gather_results(mocked_event):

    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()
    # Mock out some extra columns that would be produced by the manager's _prepare_population() method
    population["current_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12)
    population["event_step_size"] = timedelta(days=28)
    population["event_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12) + timedelta(
        days=28
    )

    lifecycle_phase = "collect_metrics"
    pop_filter = "house=='hufflepuff'"
    included_cols = ["event_time", "familiar", "house"]
    ctx.register_observation(
        observation_type=ConcatenatingObservation,
        name="foo",
        pop_filter=pop_filter,
        when=lifecycle_phase,
        included_columns=included_cols,
        results_formatter=lambda _, __: pd.DataFrame(),
    )

    filtered_pop = population.query(pop_filter)

    i = 0
    for result, _measure, _updater in ctx.gather_results(
        population, lifecycle_phase, mocked_event
    ):
        assert result.equals(filtered_pop[included_cols])
        i += 1
    assert i == 1


@pytest.mark.parametrize(
    "name, pop_filter, aggregator_sources, aggregator, stratifications",
    [
        ("wizard_count", "tracked==True", None, len, ["house", "familiar"]),
        ("power_level_total", "tracked==True", ["power_level"], sum, ["house", "familiar"]),
        (
            "wizard_time",
            "tracked==True",
            [],
            _aggregate_state_person_time,
            ["house", "familiar"],
        ),
        ("wizard_count", "tracked==True", None, len, ["familiar"]),
        ("power_level_total", "tracked==True", ["power_level"], sum, ["familiar"]),
        (
            "wizard_time",
            "tracked==True",
            [],
            _aggregate_state_person_time,
            ["familiar"],
        ),
    ],
    ids=[
        "len_aggregator_two_stratifications",
        "sum_aggregator_two_stratifications",
        "custom_aggregator_two_stratifications",
        "len_aggregator_one_stratification",
        "sum_aggregator_one_stratification",
        "custom_aggregator_one_stratification",
    ],
)
def test_gather_results_partial_stratifications_in_results(
    name, pop_filter, aggregator_sources, aggregator, stratifications, mocked_event
):
    """Test cases where not all stratifications are observed for gather_results. This looks for existence of
    unobserved stratifications and ensures their values are 0"""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()

    # Mock out some extra columns that would be produced by the manager's _prepare_population() method
    population["current_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12)
    population["event_step_size"] = timedelta(days=28)
    population["event_time"] = pd.Timestamp(year=2045, month=1, day=1, hour=12) + timedelta(
        days=28
    )
    # Remove an entire category from a stratification
    population = population[population["familiar"] != "unladen_swallow"].reset_index()

    lifecycle_phase = "collect_metrics"

    # Set up stratifications
    if "house" in stratifications:
        ctx.add_stratification(
            name="house",
            sources=["house"],
            categories=HOUSE_CATEGORIES,
            excluded_categories=None,
            mapper=None,
            is_vectorized=True,
        )
    if "familiar" in stratifications:
        ctx.add_stratification(
            name="familiar",
            sources=["familiar"],
            categories=FAMILIARS,
            excluded_categories=None,
            mapper=None,
            is_vectorized=True,
        )

    ctx.register_observation(
        observation_type=AddingObservation,
        name=name,
        pop_filter=pop_filter,
        aggregator_sources=aggregator_sources,
        aggregator=aggregator,
        stratifications=tuple(stratifications),
        when=lifecycle_phase,
        results_formatter=lambda: None,
    )

    for results, _measure, _formatter in ctx.gather_results(
        population, lifecycle_phase, mocked_event
    ):
        unladen_results = results.reset_index().query('familiar=="unladen_swallow"')
        assert len(unladen_results) > 0
        assert (unladen_results[VALUE_COLUMN] == 0).all()


def test_gather_results_with_empty_pop_filter(mocked_event):
    """Test case where pop_filter filters to an empty population. gather_results
    should return None.
    """
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()

    lifecycle_phase = "collect_metrics"
    ctx.register_observation(
        observation_type=AddingObservation,
        name="wizard_count",
        pop_filter="house == 'durmstrang'",
        aggregator_sources=[],
        aggregator=len,
        stratifications=tuple(),
        when=lifecycle_phase,
        results_formatter=lambda: None,
    )

    for result, _measure, _updater in ctx.gather_results(
        population, lifecycle_phase, mocked_event
    ):
        assert not result


def test_gather_results_with_no_stratifications(mocked_event):
    """Test case where we have no stratifications. gather_results should return one value."""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()

    lifecycle_phase = "collect_metrics"
    ctx.register_observation(
        observation_type=AddingObservation,
        name="wizard_count",
        pop_filter="",
        aggregator_sources=None,
        aggregator=len,
        stratifications=tuple(),
        when=lifecycle_phase,
        results_formatter=lambda: None,
    )

    assert len(ctx.stratifications) == 0
    assert (
        len(
            list(
                result
                for result, _measure, _updater in ctx.gather_results(
                    population, lifecycle_phase, mocked_event
                )
            )
        )
        == 1
    )


def test_bad_aggregator_stratification(mocked_event):
    """Test if an exception gets raised when a stratification that doesn't
    exist is attempted to be used, as expected."""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()
    lifecycle_phase = "collect_metrics"

    # Set up stratifications
    ctx.add_stratification(
        name="house",
        sources=["house"],
        categories=HOUSE_CATEGORIES,
        excluded_categories=None,
        mapper=None,
        is_vectorized=True,
    )
    ctx.add_stratification(
        name="familiar",
        sources=["familiar"],
        categories=FAMILIARS,
        excluded_categories=None,
        mapper=None,
        is_vectorized=True,
    )
    ctx.register_observation(
        observation_type=AddingObservation,
        name="this_shouldnt_work",
        pop_filter="",
        aggregator_sources=[],
        aggregator=sum,
        stratifications=("house", "height"),  # `height` is not a stratification
        when=lifecycle_phase,
        results_formatter=lambda: None,
    )

    with pytest.raises(KeyError, match="height"):
        for result, _measure, _updater in ctx.gather_results(
            population, lifecycle_phase, mocked_event
        ):
            print(result)


@pytest.mark.parametrize(
    "pop_filter, stratifications",
    [
        ('familiar=="cat"', tuple()),
        ('familiar=="spaghetti_yeti"', tuple()),
        ("", ("new_col1",)),
        ("", ("new_col1", "new_col2")),
        ('familiar=="cat"', ("new_col1",)),
        ("", tuple()),
    ],
    ids=[
        "pop_filter",
        "pop_filter_empties_dataframe",
        "single_excluded_stratification",
        "two_excluded_stratifications",
        "pop_filter_and_excluded_stratification",
        "no_pop_filter_or_excluded_stratifications",
    ],
)
def test__filter_population(pop_filter, stratifications):
    population = BASE_POPULATION.copy()
    if stratifications:
        # Make some of the stratifications missing to mimic mapping to excluded categories
        population["new_col1"] = "new_value1"
        population.loc[population["tracked"] == True, "new_col1"] = np.nan
        if len(stratifications) == 2:
            population["new_col2"] = "new_value2"
            population.loc[population["new_col1"].notna(), "new_col2"] = np.nan
        # Add on the post-stratified columns
        for stratification in stratifications:
            mapped_col = f"{stratification}_mapped_values"
            population[mapped_col] = population[stratification]

    filtered_pop = ResultsContext()._filter_population(
        population=population, pop_filter=pop_filter, stratification_names=stratifications
    )
    expected = population.copy()
    if pop_filter:
        familiar = pop_filter.split("==")[1].strip('"')
        expected = expected[expected["familiar"] == familiar]
    for stratification in stratifications:
        expected = expected[expected[stratification].notna()]
    assert filtered_pop.equals(expected)


@pytest.mark.parametrize(
    "stratifications, values",
    [
        (("familiar",), [FAMILIARS]),
        (("familiar", "house"), [FAMILIARS, HOUSE_CATEGORIES]),
        ((), "foo"),
    ],
)
def test__get_groups(stratifications, values):

    filtered_pop = BASE_POPULATION.copy()
    # Generate the post-stratified columns
    for stratification in stratifications:
        mapped_col = f"{stratification}_mapped_values"
        filtered_pop[mapped_col] = filtered_pop[stratification]
    groups = ResultsContext()._get_groups(
        stratifications=stratifications, filtered_pop=filtered_pop
    )
    assert isinstance(groups, DataFrameGroupBy)
    if stratifications:
        combinations = set(itertools.product(*values))
        if len(values) == 1:
            # convert from set of tuples to set of strings
            combinations = set(comb[0] for comb in combinations)
        # Check that all familiars exist
        assert set(groups.groups.keys()) == combinations
        # Check that the entire population is included
        assert sum([len(value) for value in groups.groups.values()]) == len(BASE_POPULATION)
    else:
        item = groups.groups.popitem()
        # Check that there are no other groups
        assert not groups.groups
        # Check that the group is 'all' and includes the entire population
        key, val = item
        assert key == "all"
        assert val.equals(BASE_POPULATION.index)


def test_to_observe(mocked_event, mocker):
    """Test that to_observe can be used to turn off observations"""
    ctx = ResultsContext()

    # Generate population DataFrame
    population = BASE_POPULATION.copy()

    lifecycle_phase = "collect_metrics"
    ctx.register_observation(
        observation_type=AddingObservation,
        name="wizard_count",
        pop_filter="house == 'hufflepuff'",
        aggregator_sources=[],
        aggregator=len,
        stratifications=tuple(),
        when=lifecycle_phase,
        results_formatter=lambda: None,
    )

    for result, _measure, _updater in ctx.gather_results(
        population, lifecycle_phase, mocked_event
    ):
        assert not result.empty

    # Extract the observation from the context and patch it to not observe
    observation = list(ctx.observations["collect_metrics"].values())[0][0]
    mocker.patch.object(observation, "to_observe", return_value=False)
    for result, _measure, _updater in ctx.gather_results(
        population, lifecycle_phase, mocked_event
    ):
        assert not result



# mypy: ignore-errors
"""
==================
Randomness Streams
==================

This module provides a wrapper around numpy's randomness system with the intent of coupling
it to vivarium's tools for Common Random Number genereration.


Attributes
----------
RESIDUAL_CHOICE : object
    A probability placeholder to be used in an un-normalized array of weights
    to absorb leftover weight so that the array sums to unity.
    For example::

        [0.2, 0.2, RESIDUAL_CHOICE] => [0.2, 0.2, 0.6]

        
Notes
-----
Currently this object is only used in the `choice` function of this
module.

"""

import hashlib
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from scipy import stats

from vivarium.framework.randomness.exceptions import RandomnessError
from vivarium.framework.randomness.index_map import IndexMap
from vivarium.framework.utilities import rate_to_probability

RESIDUAL_CHOICE = object()


def get_hash(key: str) -> int:
    """Gets a hash of the provided key.

    Parameters
    ----------
    key :
        A string used to create a seed for the random number generator.

    Returns
    -------
        A hash of the provided key.
    """
    max_allowable_numpy_seed = 4294967295  # 2**32 - 1
    return int(hashlib.sha1(key.encode("utf8")).hexdigest(), 16) % max_allowable_numpy_seed


class RandomnessStream:
    """A stream for producing common random numbers.

    `RandomnessStream` objects provide an interface to Vivarium's
    common random number generation.  They provide a number of methods
    for doing common simulation tasks that require random numbers like
    making decisions among a number of choices.

    Attributes
    ----------
    key
        The name of the randomness stream.
    clock
        A way to get the current simulation time.
    seed
        An extra number used to seed the random number generation.
    index_map
        A key-index mapping with a fectorized hash and vectorized lookups.
    initializes_crn_attributes
        A boolean indicating whether the stram is used to initialize CRN attributes.

    Notes
    -----
    Should not be constructed by client code.

    Simulation components get `RandomnessStream` objects by requesting
    them from the builder provided to them during the setup phase.
    I.E.::

        class VivariumComponent:
            def setup(self, builder):
                self.randomness_stream = builder.randomness.get_stream('stream_name')

    """

    def __init__(
        self,
        key: str,
        clock: Callable,
        seed: Any,
        index_map: IndexMap,
        initializes_crn_attributes: bool = False,
    ):
        self.key = key
        self.clock = clock
        self.seed = seed
        self.index_map = index_map
        self.initializes_crn_attributes = initializes_crn_attributes

    @property
    def name(self):
        return f"randomness_stream_{self.key}"

    def _key(self, additional_key: Any = None) -> str:
        """Construct a hashable key from this object's state.

        Parameters
        ----------
        additional_key
            Any additional information used to seed random number generation.

        Returns
        -------
            A key to seed random number generation.
        """
        return "_".join([self.key, str(self.clock()), str(additional_key), str(self.seed)])

    def get_draw(self, index: pd.Index, additional_key: Any = None) -> pd.Series:
        """Get an indexed set of numbers uniformly drawn from the unit interval.

        Parameters
        ----------
        index
            An index whose length is the number of random draws made
            and which indexes the returned `pandas.Series`.
        additional_key
            Any additional information used to seed random number generation.

        Returns
        -------
            A series of random numbers indexed by the provided `pandas.Index`.

        Notes
        -----
        This is the core of the CRN implementation, allowing for consistent use of random
        numbers across simulations with multiple scenarios.

        See Also
        --------
        https://en.wikipedia.org/wiki/Variance_reduction and
        "Untangling Uncertainty with Common Random Numbers:
        A Simulation Study; A.Flaxman, et. al., Summersim 2017"
        """
        # Return a structured null value if an empty index is passed
        if index.empty:
            return pd.Series(index=index, dtype=float)

        # Initialize a random_state with a seed based on the simulation clock, the
        # decision_point this stream represents, and any additional user-supplied information.
        # This is one pre-condition to reproducibility.
        seed = get_hash(self._key(additional_key))
        random_state = np.random.RandomState(seed=seed)

        # Second, we need to sample a very large chunk of random numbers. The size of the
        # index map is set at the simulation start and is at least 10x the size of the
        # initial population. Which means this is a consistently sampled block of uniformly
        # distributed random numbers, irrespective of the size of the simulation population,
        # which is important if there are scenarios that result in different population sizes
        # through time.
        sample_size = len(self.index_map)
        raw_draws = random_state.random_sample(sample_size)

        if self.initializes_crn_attributes:
            # If we're initializing CRN attributes (i.e. attributes used to identify a
            # simulant across multiple simulations), we can't use the index map yet since
            # these people aren't registered in the index map yet. Instead, we use the draws
            # from our random sample in order. This couples the initialization of CRN
            # attributes to the time step on which they are initialized, meaning interventions
            # that alter the entrance time of a simulant will break the CRN guarantees.
            draws = pd.Series(raw_draws[: len(index)], index=index)
        else:
            # If we're not initializing CRN attributes, we can use the index map to get the
            # correct draws for each simulant. This allows us to use the same CRN attributes
            # across multiple simulations, even if the population size changes.
            draw_index = self.index_map[index]
            draws = pd.Series(raw_draws[draw_index], index=index)

        return draws

    def filter_for_rate(
        self,
        population: Union[pd.DataFrame, pd.Series, pd.Index],
        rate: Union[float, List, Tuple, np.ndarray, pd.Series],
        additional_key: Any = None,
    ) -> Union[pd.DataFrame, pd.Series, pd.Index]:
        """Decide an event outcome for each individual from rates.

        Given a population or its index and an array of associated rates for
        some event to happen, we create and return the subpopulation for whom
        the event occurred.

        Parameters
        ----------
        population
            A view on the simulants for which we are determining the
            outcome of an event.
        rate
            A scalar float value or a 1d list of rates of the event under
            consideration occurring which corresponds (i.e.
            `len(population) == len(probability))` to the population view passed
            in. The rates must be scaled to the simulation time-step size either
            manually or as a post-processing step in a rate pipeline. If a
            scalar is provided, it is applied to every row in the population.
        additional_key
            Any additional information used to create the seed.

        Returns
        -------
            The subpopulation of the simulants for whom the event occurred.
            The return type will be the same as type(population).
        """
        return self.filter_for_probability(
            population, rate_to_probability(rate), additional_key
        )

    def filter_for_probability(
        self,
        population: Union[pd.DataFrame, pd.Series, pd.Index],
        probability: Union[float, List, Tuple, np.ndarray, pd.Series],
        additional_key: Any = None,
    ) -> Union[pd.DataFrame, pd.Series, pd.Index]:
        """Decide an outcome for each individual from probabilities.

        Given a population or its index and an array of associated probabilities
        for some event to happen, we create and return the subpopulation for
        whom the event occurred.

        Parameters
        ----------
        population
            A view on the simulants for which we are determining the
            outcome of an event.
        probability
            A scalar float value or a 1d list of probabilities of the event
            under consideration occurring which corresponds (i.e.
            `len(population) == len(probability)` to the population view
            passed in. If a scalar is provided, it is applied to every row in
            the population.
        additional_key
            Any additional information used to create the seed.

        Returns
        -------
            The subpopulation of the simulants for whom the event occurred.
            The return type will be the same as type(population).
        """
        if population.empty:
            return population

        index = population if isinstance(population, pd.Index) else population.index
        draws = self.get_draw(index, additional_key)
        mask = np.array(draws < probability)
        return population[mask]

    def choice(
        self,
        index: pd.Index,
        choices: Union[List, Tuple, np.ndarray, pd.Series],
        p: Optional[Union[List, Tuple, np.ndarray, pd.Series]] = None,
        additional_key: Optional[Any] = None,
    ) -> pd.Series:
        """Decides between a weighted or unweighted set of choices.

        Given a set of choices with or without corresponding weights,
        returns an indexed set of decisions from those choices. This is
        simply a vectorized way to make decisions with some book-keeping.

        Parameters
        ----------
        index
            An index whose length is the number of random draws made
            and which indexes the returned `pandas.Series`.
        choices
            A set of options to choose from.
        p
            The relative weights of the choices.  Can be either a 1-d array of
            the same length as `choices` or a 2-d array with `len(index)` rows
            and `len(choices)` columns.  In the 1-d case, the same set of
            weights are used to decide among the choices for every item in
            the `index`. In the 2-d case, each row in `p` contains a separate
            set of weights for every item in the `index`.
        additional_key
            Any additional information used to seed random number generation.

        Returns
        -------
            An indexed set of decisions from among the available `choices`.

        Raises
        ------
        RandomnessError
            If any row in `p` contains `RESIDUAL_CHOICE` and the remaining
            weights in the row are not normalized or any row of `p contains
            more than one reference to `RESIDUAL_CHOICE`.
        """
        draws = self.get_draw(index, additional_key)
        return _choice(draws, choices, p)

    def sample_from_distribution(
        self,
        index: pd.Index,
        distribution: Optional[stats.rv_continuous] = None,
        ppf: Optional[Callable[[pd.Series, dict[str, Any]], pd.Series]] = None,
        additional_key: Any = None,
        **distribution_kwargs: dict[str, Any],
    ) -> pd.Series:
        """Given a distribution, returns an indexed set of samples from that
        distribution.

        Parameters
        ----------
        index
            An index whose length is the number of random draws made
            and which indexes the returned `pandas.Series`.
        distribution
            A scipy.stats distribution object.
        ppf
            A function that takes a series of draws and returns a series of samples.
        additional_key
            Any additional information used to seed random number generation.
        distribution_kwargs
            Additional keyword arguments to pass to the distribution's ppf function.

        Returns
        -------
            An indexed set of samples from the provided distribution.
        """
        if ppf is None and distribution is None:
            raise ValueError("Either distribution or ppf must be provided")
        if ppf is not None and distribution is not None:
            raise ValueError("Only one of distribution or ppf can be provided")

        if distribution is not None:
            ppf = distribution.ppf

        draws = self.get_draw(index, additional_key)
        samples = pd.Series(ppf(draws, **distribution_kwargs), index=index)
        return samples

    def __repr__(self) -> str:
        return "RandomnessStream(key={!r}, clock={!r}, seed={!r})".format(
            self.key, self.clock(), self.seed
        )


def _choice(
    draws: pd.Series,
    choices: Union[List, Tuple, np.ndarray, pd.Series],
    p: Optional[Union[List, Tuple, np.ndarray, pd.Series]] = None,
) -> pd.Series:
    """Decides between a weighted or unweighted set of choices.

    Given a set of choices with or without corresponding weights,
    returns an indexed set of decisions from those choices. This is
    simply a vectorized way to make decisions with some book-keeping.

    Parameters
    ----------
    draws
        A uniformly distributed random number for every simulant to make
        a choice for.
    choices
        A set of options to choose from. Choices must be the same for every
        simulant.
    p
        The relative weights of the choices.  Can be either a 1-d array of
        the same length as `choices` or a 2-d array with `len(draws)` rows
        and `len(choices)` columns.  In the 1-d case, the same set of weights
        are used to decide among the choices for every item in the `index`.
        In the 2-d case, each row in `p` contains a separate set of weights
        for every item in the `index`.

    Returns
    -------
        An indexed set of decisions from among the available `choices`.

    Raises
    ------
    RandomnessError
        If any row in `p` contains `RESIDUAL_CHOICE` and the remaining
        weights in the row are not normalized or any row of `p` contains
        more than one reference to `RESIDUAL_CHOICE`.
    """
    # Convert p to normalized probabilities broadcasted over index.
    p = (
        _set_residual_probability(_normalize_shape(p, draws.index))
        if p is not None
        else np.ones((len(draws.index), len(choices)))
    )
    p = p / p.sum(axis=1, keepdims=True)

    p_bins = np.cumsum(p, axis=1)
    # Use the random draw to make a choice for every row in index.
    choice_index = (draws.values[np.newaxis].T > p_bins).sum(axis=1)

    return pd.Series(np.array(choices)[choice_index], index=draws.index)


def _normalize_shape(
    p: Union[List, Tuple, np.ndarray, pd.Series],
    index: Union[pd.Index, pd.MultiIndex],
) -> np.ndarray:
    p = np.array(p)
    # We got a 1-d array => same weights for every index.
    if len(p.shape) == 1:
        # Turn 1-d array into 2-d array with same weights in every row.
        p = np.array(np.broadcast_to(p, (len(index), p.shape[0])))
    return p


def _set_residual_probability(p: np.ndarray) -> np.ndarray:
    """Turns any use of `RESIDUAL_CHOICE` into a residual probability.

    Parameters
    ----------
    p
        Array where each row is a set of probability weights and potentially
        a `RESIDUAL_CHOICE` placeholder.

    Returns
    -------
        Array where each row is a set of normalized probability weights.

    Raises
    ------
    RandomnessError
        If more than one residual choice is supplied for a single set of weights.
    RandomnessError
        If residual choice is supplied with weights that sum to more than 1.
    """
    residual_mask = p == RESIDUAL_CHOICE
    if residual_mask.any():  # I.E. if we have any placeholders.
        if np.any(np.sum(residual_mask, axis=1) - 1):
            msg = (
                "More than one residual choice supplied for a single "
                f"set of weights. Weights: {p}."
            )
            raise RandomnessError(msg)

        p[residual_mask] = 0
        residual_p = 1 - np.sum(p, axis=1)  # Probabilities sum to 1.

        if np.any(residual_p < 0):  # We got un-normalized probability weights.
            msg = (
                "Residual choice supplied with weights that summed to more than 1. "
                f"Weights: {p}."
            )
            raise RandomnessError(msg)

        p[residual_mask] = residual_p
    return p



import pytest

from vivarium import Component
from vivarium.framework.population.exceptions import PopulationError
from vivarium.framework.population.manager import InitializerComponentSet, PopulationManager


def test_initializer_set_fail_type():
    component_set = InitializerComponentSet()

    with pytest.raises(TypeError):
        component_set.add(lambda: "test", ["test_column"])

    def initializer():
        return "test"

    with pytest.raises(TypeError):
        component_set.add(initializer, ["test_column"])


class NonComponent:
    def initializer(self) -> str:
        return "test"


class InitializingComponent(Component):
    @property
    def name(self) -> str:
        return self._name

    def __init__(self, name: str):
        super().__init__()
        self._name = name

    def initializer(self) -> str:
        return "test"

    def other_initializer(self) -> str:
        return "whoops"


def test_initializer_set_fail_attr():
    component_set = InitializerComponentSet()

    with pytest.raises(AttributeError):
        component_set.add(NonComponent().initializer, ["test_column"])


def test_initializer_set_duplicate_component():
    component_set = InitializerComponentSet()
    component = InitializingComponent("test")

    component_set.add(component.initializer, ["test_column1"])
    with pytest.raises(PopulationError, match="multiple population initializers"):
        component_set.add(component.other_initializer, ["test_column2"])


def test_initializer_set_duplicate_columns():
    component_set = InitializerComponentSet()
    component1 = InitializingComponent("test1")
    component2 = InitializingComponent("test2")
    columns = ["test_column"]

    component_set.add(component1.initializer, columns)
    with pytest.raises(PopulationError, match="both registered initializers"):
        component_set.add(component2.initializer, columns)

    with pytest.raises(PopulationError, match="both registered initializers"):
        component_set.add(component2.initializer, ["sneaky_column"] + columns)


def test_initializer_set_population_manager():
    component_set = InitializerComponentSet()
    population_manager = PopulationManager()

    component_set.add(population_manager.on_initialize_simulants, ["tracked"])


def test_initializer_set():
    component_set = InitializerComponentSet()
    for i in range(10):
        component = InitializingComponent(str(i))
        columns = [f"test_column_{i}_{j}" for j in range(5)]
        component_set.add(component.initializer, columns)


def test_get_view_with_no_query():
    manager = PopulationManager()
    view = manager._get_view(columns=["age", "sex"])
    assert view.query == "tracked == True"



from typing import Any, Dict

from vivarium import Component
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event


class MockComponentA(Component):
    @property
    def name(self) -> str:
        return self._name

    def __init__(self, *args, name="mock_component_a"):
        super().__init__()
        self._name = name
        self.args = args
        self.builder_used_for_setup = None

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.name == other.name


class MockComponentB(Component):
    @property
    def name(self) -> str:
        return self._name

    def __init__(self, *args, name="mock_component_b"):
        super().__init__()
        self._name = name
        self.args = args
        self.builder_used_for_setup = None
        if len(self.args) > 1:
            for arg in self.args:
                self._sub_components.append(MockComponentB(arg, name=arg))

    def setup(self, builder: Builder) -> None:
        self.builder_used_for_setup = builder
        builder.value.register_value_modifier("metrics", self.metrics)

    def metrics(self, _, metrics):
        if "test" in metrics:
            metrics["test"] += 1
        else:
            metrics["test"] = 1
        return metrics

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.name == other.name


class MockGenericComponent(Component):
    CONFIGURATION_DEFAULTS = {
        "component": {
            "key1": "val",
            "key2": {
                "subkey1": "val",
                "subkey2": "val",
            },
            "key3": ["val", "val", "val"],
        }
    }

    @property
    def name(self) -> str:
        return self._name

    @property
    def configuration_defaults(self) -> Dict[str, Any]:
        return {self.name: self.CONFIGURATION_DEFAULTS["component"]}

    def __init__(self, name: str):
        super().__init__()
        self._name = name
        self.builder_used_for_setup = None

    def setup(self, builder: Builder) -> None:
        self.builder_used_for_setup = builder
        self.config = builder.configuration[self.name]

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.name == other.name


class Listener(MockComponentB):
    def __init__(self, *args, name="test_listener"):
        super().__init__(*args, name=name)
        self.post_setup_called = False
        self.time_step_prepare_called = False
        self.time_step_called = False
        self.time_step_cleanup_called = False
        self.collect_metrics_called = False
        self.simulation_end_called = False

        self.event_indexes = {
            "time_step_prepare": None,
            "time_step": None,
            "time_step_cleanup": None,
            "collect_metrics": None,
        }

    def on_post_setup(self, event: Event) -> None:
        self.post_setup_called = True

    def on_time_step_prepare(self, event: Event) -> None:
        self.time_step_prepare_called = True
        self.event_indexes["time_step_prepare"] = event.index

    def on_time_step(self, event: Event) -> None:
        self.time_step_called = True
        self.event_indexes["time_step"] = event.index

    def on_time_step_cleanup(self, event: Event) -> None:
        self.time_step_cleanup_called = True
        self.event_indexes["time_step_cleanup"] = event.index

    def on_collect_metrics(self, event: Event) -> None:
        self.collect_metrics_called = True
        self.event_indexes["collect_metrics"] = event.index

    def on_simulation_end(self, event: Event) -> None:
        self.simulation_end_called = True



# mypy: ignore-errors
"""
=========
Component
=========

A base Component class to be used to create components for use in ``vivarium``
simulations.

"""

import re
from abc import ABC
from importlib import import_module
from inspect import signature
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Union

import pandas as pd
from layered_config_tree import ConfigurationError, LayeredConfigTree
from loguru._logger import Logger

from vivarium.framework.artifact import ArtifactException
from vivarium.framework.event import Event
from vivarium.framework.lookup import LookupTable
from vivarium.framework.population import PopulationError, PopulationView

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder
    from vivarium.framework.population import SimulantData

DEFAULT_EVENT_PRIORITY = 5
"""The default priority at which events will be triggered."""


class Component(ABC):
    """The base class for all components used in a Vivarium simulation.

    A `Component` in a Vivarium simulation represents a distinct feature or
    aspect of the model. It encapsulates the logic and data needed for that
    feature. Components commonly interact with the rest of the simulation by
    creating and updating columns in the state table, registering pipelines,
    and registering modifiers on pipelines created by other components. Observer
    components might also register observations. All components within a
    simulation must have a unique name, which is generated by default from the
    component's class and the argument passed to its constructor.

    The `setup_component` is run by Vivarium during the setup phase and performs
    a series of operations to prepare the component for the simulation. These
    operations include setting the logger for the component, calling the
    component's custom `setup` method, setting the population view if the
    component needs one, and registering listeners for each lifecycle event if
    the component has defined a method to be triggered on that event.

    Subclasses of `Component` should override these properties as needed:

    - `sub_components`
    - `configuration_defaults`
    - `columns_created`
    - `columns_required`
    - `initialization_requirements`
    - `population_view_query`
    - `post_setup_priority`
    - `time_step_prepare_priority`
    - `time_step_priority`
    - `time_step_cleanup_priority`
    - `collect_metrics_priority`
    - `simulation_end_priority`

    Subclasses of `Component` should override these methods in order to have
    operations occur during the appropriate lifecycle phase of a simulation:

    - `setup`
    - `on_post_setup`
    - `on_initialize_simulants`
    - `on_time_step_prepare`
    - `on_time_step`
    - `on_time_step_cleanup`
    - `on_collect_metrics`
    - `on_simulation_end`

    """

    CONFIGURATION_DEFAULTS: Dict[str, Any] = {}
    """A dictionary containing the defaults for any configurations managed by this
    component. An empty dictionary indicates no managed configurations.
    """

    def __repr__(self):
        """Returns a string representation of the __init__ call made to create this
        object.

        The representation is built by retrieving the initialization parameters
        and their values. If a value is an instance of Component, its own
        __repr__() is called. The resulting string is stored in the _repr
        attribute and returned.

        IMPORTANT: this method must not be called within the `__init__`
        functions of this component or its subclasses or its value may not be
        initialized correctly.

        Returns
        -------
            A string representation of the __init__ call made to create this
            object.
        """
        if not self._repr:
            args = [
                f"{name}={value.__repr__() if isinstance(value, Component) else value}"
                for name, value in self.get_initialization_parameters().items()
            ]
            args = ", ".join(args)
            self._repr = f"{type(self).__name__}({args})"

        return self._repr

    def __str__(self):
        return self._repr

    ##############
    # Properties #
    ##############

    @property
    def name(self) -> str:
        """Returns the name of the component.

        By convention, these are in snake case with arguments of the `__init__`
        appended and separated by `.`.

        Names must be unique within a simulation.

        The name is created by first converting the name of the class to snake
        case. Then, the names of the initialization parameters are appended,
        separated by `.`. If a parameter is an instance of Component, its name
        property is used; otherwise, the string representation of the parameter
        is used. The resulting string is stored in the _name attribute and
        returned.

        IMPORTANT: this property must not be accessed within the `__init__`
        functions of this component or its subclasses or its value may not be
        initialized correctly.

        Returns
        -------
        str
            The name of the component.
        """
        if not self._name:
            base_name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", type(self).__name__)
            base_name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", base_name).lower()

            args = [
                f"'{value.name}'" if isinstance(value, Component) else str(value)
                for value in self.get_initialization_parameters().values()
            ]
            self._name = ".".join([base_name] + args)

        return self._name

    @property
    def population_view(self) -> PopulationView:
        """Provides the PopulationView for this component.

        Returns
        -------
        PopulationView
            The PopulationView for this component

        Raises
        ------
        PopulationError
            If the component does not have access to the state table.
        """
        if self._population_view is None:
            raise PopulationError(
                f"Component '{self.name}' does not have access to the state "
                "table. This is likely due to a failure to set columns_required "
                "or columns_created for this component."
            )
        return self._population_view

    @property
    def sub_components(self) -> List["Component"]:
        """Provide components managed by this component.

        Returns
        -------
        List[Component]
            The sub-components that are managed by this component.
        """
        return self._sub_components

    @property
    def configuration_defaults(self) -> Dict[str, Any]:
        """Provides a dictionary containing the defaults for any configurations
        managed by this component.

        These default values will be stored at the `component_configs` layer of the
        simulation's LayeredConfigTree.

        Returns
        -------
            A dictionary containing the defaults for any configurations managed by
            this component.
        """
        return self.CONFIGURATION_DEFAULTS

    @property
    def columns_created(self) -> List[str]:
        """Provides names of columns created by the component.

        Returns
        -------
            Names of the columns created by this component, or an empty list if
            none.
        """
        return []

    @property
    def columns_required(self) -> Optional[List[str]]:
        """Provides names of columns required by the component.

        Returns
        -------
            Names of required columns not created by this component. An empty
            list means all available columns are needed. `None` means no
            additional columns are necessary.
        """
        return None

    @property
    def initialization_requirements(self) -> Dict[str, List[str]]:
        """Provides the names of all values required by this component during
        simulant initialization.

        Returns
        -------
            A dictionary containing the additional requirements of this
            component during simulant initialization. An omitted key or an empty
            list for a key implies no requirements for that key during
            initialization.
        """
        return {
            "requires_columns": [],
            "requires_values": [],
            "requires_streams": [],
        }

    @property
    def population_view_query(self) -> Optional[str]:
        """Provides a query to use when filtering the component's `PopulationView`.

        Returns
        -------
            A pandas query string for filtering the component's `PopulationView`.
            Returns `None` if no filtering is required.
        """
        return None

    @property
    def post_setup_priority(self) -> int:
        """Provides the priority of this component's post_setup listener.

        Returns
        -------
            The priority of this component's post_setup listener. This value
            can range from 0 to 9, inclusive.
        """
        return DEFAULT_EVENT_PRIORITY

    @property
    def time_step_prepare_priority(self) -> int:
        """Provides the priority of this component's time_step__prepare listener.

        Returns
        -------
            The priority of this component's time_step__prepare listener. This value
            can range from 0 to 9, inclusive.
        """
        return DEFAULT_EVENT_PRIORITY

    @property
    def time_step_priority(self) -> int:
        """Provides the priority of this component's time_step listener.

        Returns
        -------
            The priority of this component's time_step listener. This value
            can range from 0 to 9, inclusive.
        """
        return DEFAULT_EVENT_PRIORITY

    @property
    def time_step_cleanup_priority(self) -> int:
        """Provides the priority of this component's time_step__cleanup listener.

        Returns
        -------
            The priority of this component's time_step__cleanup listener. This value
            can range from 0 to 9, inclusive.
        """
        return DEFAULT_EVENT_PRIORITY

    @property
    def collect_metrics_priority(self) -> int:
        """Provides the priority of this component's collect_metrics listener.

        Returns
        -------
            The priority of this component's collect_metrics listener. This value
            can range from 0 to 9, inclusive.
        """
        return DEFAULT_EVENT_PRIORITY

    @property
    def simulation_end_priority(self) -> int:
        """Provides the priority of this component's simulation_end listener.

        Returns
        -------
            The priority of this component's simulation_end listener. This value
            can range from 0 to 9, inclusive.
        """
        return DEFAULT_EVENT_PRIORITY

    #####################
    # Lifecycle methods #
    #####################

    def __init__(self) -> None:
        """Initializes a new instance of the Component class.

        This method is the initializer for the Component class. It initializes
        logger of type Logger and population_view of type PopulationView to None.
        These attributes will be fully initialized in the setup_component method
        of this class.
        """
        self._repr: str = ""
        self._name: str = ""
        self._sub_components: List["Component"] = []
        self.logger: Optional[Logger] = None
        self.get_value_columns: Optional[
            Callable[[Union[str, pd.DataFrame]], List[str]]
        ] = None
        self.configuration: Optional[LayeredConfigTree] = None
        self._population_view: Optional[PopulationView] = None
        self.lookup_tables: Dict[str, LookupTable] = {}

    def setup_component(self, builder: "Builder") -> None:
        """Sets up the component for a Vivarium simulation.

        This method is run by Vivarium during the setup phase. It performs a series
        of operations to prepare the component for the simulation.

        It sets the logger for the component, sets up the component, sets the
        population view, and registers various listeners including post_setup,
        simulant_initializer, time_step_prepare, time_step, time_step_cleanup,
        collect_metrics, and simulation_end listeners.

        Parameters
        ----------
        builder
            The builder object used to set up the component.
        """
        self.logger = builder.logging.get_logger(self.name)
        self.get_value_columns = builder.data.value_columns()
        self.configuration = self.get_configuration(builder)
        self.build_all_lookup_tables(builder)
        self.setup(builder)
        self._set_population_view(builder)
        self._register_post_setup_listener(builder)
        self._register_simulant_initializer(builder)
        self._register_time_step_prepare_listener(builder)
        self._register_time_step_listener(builder)
        self._register_time_step_cleanup_listener(builder)
        self._register_collect_metrics_listener(builder)
        self._register_simulation_end_listener(builder)

    #######################
    # Methods to override #
    #######################

    def setup(self, builder: "Builder") -> None:
        """Defines custom actions this component needs to run during the setup
        lifecycle phase.

        This method is intended to be overridden by subclasses to perform any
        necessary setup operations specific to the component. By default, it
        does nothing.

        Parameters
        ----------
        builder
            The builder object used to set up the component.
        """
        pass

    def on_post_setup(self, event: Event) -> None:
        """Method that vivarium will run during the post_setup event.

        This method is intended to be overridden by subclasses if there are
        operations they need to perform specifically during the post_setup event.

        Notes
        -----
        This method is not commonly used functionality.

        Parameters
        ----------
        event
            The event object associated with the post_setup event.
        """
        pass

    def on_initialize_simulants(self, pop_data: "SimulantData") -> None:
        """
        Method that vivarium will run during simulant initialization.

        This method is intended to be overridden by subclasses if there are
        operations they need to perform specifically during the simulant
        initialization phase.

        Parameters
        ----------
        pop_data : SimulantData
            The data associated with the simulants being initialized.

        Returns
        -------
        None
        """
        pass

    def on_time_step_prepare(self, event: Event) -> None:
        """Method that vivarium will run during the time_step__prepare event.

        This method is intended to be overridden by subclasses if there are
        operations they need to perform specifically during the
        time_step__prepare event.

        Parameters
        ----------
        event
            The event object associated with the time_step__prepare event.
        """
        pass

    def on_time_step(self, event: Event) -> None:
        """Method that vivarium will run during the time_step event.

        This method is intended to be overridden by subclasses if there are
        operations they need to perform specifically during the time_step event.

        Parameters
        ----------
        event
            The event object associated with the time_step event.
        """
        pass

    def on_time_step_cleanup(self, event: Event) -> None:
        """Method that vivarium will run during the time_step__cleanup event.

        This method is intended to be overridden by subclasses if there are
        operations they need to perform specifically during the
        time_step__cleanup event.

        Parameters
        ----------
        event
            The event object associated with the time_step__cleanup event.
        """
        pass

    def on_collect_metrics(self, event: Event) -> None:
        """Method that vivarium will run during the collect_metrics event.

        This method is intended to be overridden by subclasses if there are
        operations they need to perform specifically during the collect_metrics
        event.

        Parameters
        ----------
        event
            The event object associated with the collect_metrics event.
        """
        pass

    def on_simulation_end(self, event: Event) -> None:
        """Method that vivarium will run during the simulation_end event.

        This method is intended to be overridden by subclasses if there are
        operations they need to perform specifically during the simulation_end
        event.

        Parameters
        ----------
        event
            The event object associated with the simulation_end event.
        """
        pass

    ##################
    # Helper methods #
    ##################

    def get_initialization_parameters(self) -> Dict[str, Any]:
        """Retrieves the values of all parameters specified in the `__init__` that
        have an attribute with the same name.

        Notes
        -----
        This retrieves the value of the attribute at the time of calling,
        which is not guaranteed to be the same as the original value.

        Returns
        -------
            A dictionary where the keys are the names of the parameters used in
            the `__init__` method and the values are their current values.
        """
        return {
            parameter_name: getattr(self, parameter_name)
            for parameter_name in signature(self.__init__).parameters
            if hasattr(self, parameter_name)
        }

    def get_configuration(self, builder: "Builder") -> Optional[LayeredConfigTree]:
        """Retrieves the configuration for this component from the builder.

        This method retrieves the configuration for this component from the
        simulation's overall configuration. The configuration is retrieved using
        the name of the component as the key.

        Parameters
        ----------
        builder
            The simulation's builder object.

        Returns
        -------
            The configuration for this component, or `None` if the component has
            no configuration.
        """

        if self.name in builder.configuration:
            return builder.configuration[self.name]
        return None

    def build_all_lookup_tables(self, builder: "Builder") -> None:
        """Builds all lookup tables for this component.

        This method builds lookup tables for this component based on the data
        sources specified in the configuration. If no data sources are specified,
        no lookup tables are built.

        The created lookup tables are stored in the lookup_tables dictionary of
        the component, with the table name as the key.

        Parameters
        ----------
        builder
            The builder object used to set up the component.
        """
        if self.configuration and "data_sources" in self.configuration:
            for table_name in self.configuration.data_sources.keys():
                try:
                    self.lookup_tables[table_name] = self.build_lookup_table(
                        builder, self.configuration.data_sources[table_name]
                    )
                except ConfigurationError as e:
                    raise ConfigurationError(
                        f"Error building lookup table '{table_name}': {e}"
                    )

    def build_lookup_table(
        self,
        builder: "Builder",
        # todo: replace with LookupTableData
        data_source: Union[str, float, int, list, pd.DataFrame],
        value_columns: Optional[Iterable[str]] = None,
    ) -> LookupTable:
        """Builds a LookupTable from a data source.

        Uses `get_data` to parse the data source and retrieve the lookup table
        data. The LookupTable is built from the data source, with the value
        columns specified in the value_columns parameter. If value_columns is
        None and the data is a DataFrame, the ArtifactManager will determine
        the value columns.

        Parameters
        ----------
        builder
            The builder object used to set up the component.
        data_source
            The data source to build the LookupTable from.
        value_columns
            The columns to include in the LookupTable.

        Returns
        -------
            The LookupTable built from the data source.

        Raises
        ------
        layered_config_tree.exceptions.ConfigurationError
            If the data source is invalid.
        """
        data = self.get_data(builder, data_source)

        if isinstance(data, list):
            return builder.lookup.build_table(data, value_columns=list(value_columns))
        if isinstance(data, pd.DataFrame):
            all_columns = set(data.columns)
            if value_columns is None:
                value_columns = set(self.get_value_columns(data))
            else:
                value_columns = set(value_columns)

            potential_parameter_columns = [
                str(col).removesuffix("_start")
                for col in all_columns
                if str(col).endswith("_start")
            ]
            parameter_columns = set()
            bin_edge_columns = set()
            for column in potential_parameter_columns:
                if f"{column}_end" in all_columns:
                    parameter_columns.add(column)
                    bin_edge_columns.update([f"{column}_start", f"{column}_end"])

            key_columns = all_columns - value_columns - bin_edge_columns

            return builder.lookup.build_table(
                data=data,
                key_columns=list(key_columns),
                parameter_columns=list(parameter_columns),
                value_columns=list(value_columns),
            )

        return builder.lookup.build_table(data)

    def get_data(
        # TODO: replace with LookupTableData
        self,
        builder: "Builder",
        data_source: Union[str, float, pd.DataFrame],
    ) -> Union[float, pd.DataFrame]:
        """Retrieves data from a data source.

        If the data source is a float or a DataFrame, it is treated as the data
        itself. If the data source is a string, containing the substring '::',
        it is treated as a function to call to retrieve the data. The string to
        the left of '::' is the module to import, and the string to the right is
        the function to call. 'self' can be provided to the left of '::' to call
        a method on the component itself. If the data source is a string without
        the substring '::', it is treated as a key in the artifact.

        Parameters
        ----------
        builder
            The builder object used to set up the component.
        data_source
            The data source to retrieve data from.

        Returns
        -------
            The data retrieved from the data source.

        Raises
        ------
        layered_config_tree.exceptions.ConfigurationError
            If the data source is invalid.
        """
        if isinstance(data_source, (float, int, list, pd.DataFrame)):
            return data_source

        if "::" in data_source:
            module, method = data_source.split("::")
            try:
                if module == "self":
                    data_getter = getattr(self, method)
                else:
                    data_getter = getattr(import_module(module), method)
            except ModuleNotFoundError:
                raise ConfigurationError(f"Unable to find module '{module}'.")
            except AttributeError:
                module_string = (
                    f"component {self.name}." if module == "self" else f"module '{module}'."
                )
                raise ConfigurationError(
                    f"There is no method '{method}' for the {module_string}."
                )

            return data_getter(builder)

        try:
            return builder.data.load(data_source)
        except ArtifactException:
            raise ConfigurationError(f"Failed to find key '{data_source}' in artifact.")

    def _set_population_view(self, builder: "Builder") -> None:
        """Creates the PopulationView for this component if it needs access to
        the state table.

        The method determines the necessary columns for the PopulationView
        based on the columns required and created by this component. If no
        columns are required or created, no PopulationView is set.

        Parameters
        ----------
        builder
            The builder object used to set up the component.
        """
        if self.columns_required:
            # Get all columns created and required
            population_view_columns = self.columns_created + self.columns_required
        elif self.columns_required == []:
            # Empty list means population view needs all available columns
            population_view_columns = []
        elif self.columns_required is None and self.columns_created:
            # No additional columns required, so just get columns created
            population_view_columns = self.columns_created
        else:
            # no need for a population view if no columns created or required
            population_view_columns = None

        if population_view_columns is not None:
            self._population_view = builder.population.get_view(
                population_view_columns, self.population_view_query
            )

    def _register_post_setup_listener(self, builder: "Builder") -> None:
        """Registers a post_setup listener if this component has defined one.

        This method allows the component to respond to "post_setup" events if it
        has its own `on_post_setup` method. The listener will be registered with
        the component's post_setup priority, allowing control over the order of
        operations when multiple components are listening to the same event.

        Parameters
        ----------
        builder
            The builder with which to register the listener.
        """
        if type(self).on_post_setup != Component.on_post_setup:
            builder.event.register_listener(
                "post_setup",
                self.on_post_setup,
                self.post_setup_priority,
            )

    def _register_simulant_initializer(self, builder: "Builder") -> None:
        """Registers a simulant initializer if this component has defined one.

        This method allows the component to initialize simulants if it has its
        own `on_initialize_simulants` method. It registers this method with the
        builder's `PopulationManager``. It also specifies the columns that the
        component creates and any additional requirements for initialization.

        Parameters
        ----------
        builder
            The builder with which to register the initializer.
        """
        if type(self).on_initialize_simulants != Component.on_initialize_simulants:
            builder.population.initializes_simulants(
                self.on_initialize_simulants,
                creates_columns=self.columns_created,
                **self.initialization_requirements,
            )

    def _register_time_step_prepare_listener(self, builder: "Builder") -> None:
        """Registers a time_step_prepare listener if this component has defined one.

        This method allows the component to respond to "time_step_prepare" events
        if it has its own `on_time_step_prepare` method. The listener will be
        registered with the component's time_step_prepare priority.

        Parameters
        ----------
        builder
            The builder with which to register the listener.
        """
        if type(self).on_time_step_prepare != Component.on_time_step_prepare:
            builder.event.register_listener(
                "time_step__prepare",
                self.on_time_step_prepare,
                self.time_step_prepare_priority,
            )

    def _register_time_step_listener(self, builder: "Builder") -> None:
        """Registers a time_step listener if this component has defined one.

        This method allows the component to respond to "time_step" events
        if it has its own `on_time_step` method. The listener will be
        registered with the component's time_step priority.

        Parameters
        ----------
        builder
            The builder with which to register the listener.
        """
        if type(self).on_time_step != Component.on_time_step:
            builder.event.register_listener(
                "time_step",
                self.on_time_step,
                self.time_step_priority,
            )

    def _register_time_step_cleanup_listener(self, builder: "Builder") -> None:
        """Registers a time_step_cleanup listener if this component has defined one.

        This method allows the component to respond to "time_step_cleanup" events
        if it has its own `on_time_step_cleanup` method. The listener will be
        registered with the component's time_step_cleanup priority.

        Parameters
        ----------
        builder
            The builder with which to register the listener.
        """
        if type(self).on_time_step_cleanup != Component.on_time_step_cleanup:
            builder.event.register_listener(
                "time_step__cleanup",
                self.on_time_step_cleanup,
                self.time_step_cleanup_priority,
            )

    def _register_collect_metrics_listener(self, builder: "Builder") -> None:
        """Registers a collect_metrics listener if this component has defined one.

        This method allows the component to respond to "collect_metrics" events
        if it has its own `on_collect_metrics` method. The listener will be
        registered with the component's collect_metrics priority.

        Parameters
        ----------
        builder
            The builder with which to register the listener.
        """
        if type(self).on_collect_metrics != Component.on_collect_metrics:
            builder.event.register_listener(
                "collect_metrics",
                self.on_collect_metrics,
                self.collect_metrics_priority,
            )

    def _register_simulation_end_listener(self, builder: "Builder") -> None:
        """Registers a simulation_end listener if this component has defined one.

        This method allows the component to respond to "simulation_end" events
        if it has its own `on_simulation_end` method. The listener will be
        registered with the component's simulation_end priority.

        Parameters
        ----------
        builder
            The builder with which to register the listener.
        """
        if type(self).on_simulation_end != Component.on_simulation_end:
            builder.event.register_listener(
                "simulation_end",
                self.on_simulation_end,
                self.simulation_end_priority,
            )



# mypy: ignore-errors
"""
===============
Stratifications
===============

"""

from dataclasses import dataclass
from typing import Callable, List, Optional, Union

import pandas as pd
from pandas.api.types import CategoricalDtype

from vivarium.types import ScalarValue

STRATIFICATION_COLUMN_SUFFIX: str = "mapped_values"


@dataclass
class Stratification:
    """Class for stratifying observed quantities by specified characteristics.

    Each Stratification represents a set of mutually exclusive and collectively
    exhaustive categories into which simulants can be assigned.

    This class includes a :meth:`stratify <stratify>` method that produces an
    output column by calling the mapper on the source columns.

    """

    name: str
    """Name of the stratification."""
    sources: List[str]
    """A list of the columns and values needed as input for the `mapper`."""
    categories: List[str]
    """Exhaustive list of all possible stratification values."""
    excluded_categories: List[str]
    """List of possible stratification values to exclude from results processing.
    If None (the default), will use exclusions as defined in the configuration."""
    mapper: Optional[
        Union[
            Callable[[Union[pd.Series, pd.DataFrame]], pd.Series],
            Callable[[ScalarValue], str],
        ]
    ]
    """A callable that maps the columns and value pipelines specified by the
    `requires_columns` and `requires_values` arguments to the stratification
    categories. It can either map the entire population or an individual
    simulant. A simulation will fail if the `mapper` ever produces an invalid
    value."""
    is_vectorized: bool = False
    """True if the `mapper` function will map the entire population, and False
    if it will only map a single simulant."""

    def __str__(self) -> str:
        return (
            f"Stratification '{self.name}' with sources {self.sources}, "
            f"categories {self.categories}, and mapper {self.mapper.__name__}"
        )

    def __post_init__(self) -> None:
        """Assign a default `mapper` if none was provided and check for non-empty
        `categories` and `sources` otherwise.

        Raises
        ------
        ValueError
            If no mapper is provided and the number of sources is not 1.
        ValueError
            If the categories argument is empty.
        ValueError
            If the sources argument is empty.
        """
        if self.mapper is None:
            if len(self.sources) != 1:
                raise ValueError(
                    f"No mapper but {len(self.sources)} stratification sources are "
                    f"provided for stratification {self.name}. The list of sources "
                    "must be of length 1 if no mapper is provided."
                )
            self.mapper = self._default_mapper
            self.is_vectorized = True
        if not self.categories:
            raise ValueError("The categories argument must be non-empty.")
        if not self.sources:
            raise ValueError("The sources argument must be non-empty.")

    def stratify(self, population: pd.DataFrame) -> pd.Series:
        """Apply the `mapper` to the population `sources` columns to create a new
        Series to be added to the population.

        Any `excluded_categories` (which have already been removed from `categories`)
        will be converted to NaNs in the new column and dropped later at the
        observation level.

        Parameters
        ----------
        population
            A DataFrame containing the data to be stratified.

        Returns
        -------
            A Series containing the mapped values to be used for stratifying.

        Raises
        ------
        ValueError
            If the mapper returns any values not in `categories` or `excluded_categories`.
        """
        if self.is_vectorized:
            mapped_column = self.mapper(population[self.sources])
        else:
            mapped_column = population[self.sources].apply(self.mapper, axis=1)
        unknown_categories = set(mapped_column) - set(
            self.categories + self.excluded_categories
        )
        # Reduce all nans to a single one
        unknown_categories = [cat for cat in unknown_categories if not pd.isna(cat)]
        if mapped_column.isna().any():
            unknown_categories.append(mapped_column[mapped_column.isna()].iat[0])
        if unknown_categories:
            raise ValueError(f"Invalid values mapped to {self.name}: {unknown_categories}")

        # Convert the dtype to the allowed categories. Note that this will
        # result in Nans for any values in excluded_categories.
        mapped_column = mapped_column.astype(
            CategoricalDtype(categories=self.categories, ordered=True)
        )
        return mapped_column

    @staticmethod
    def _default_mapper(pop: pd.DataFrame) -> pd.Series:
        """Default stratification mapper that squeezes a DataFrame to a Series.

        Parameters
        ----------
        pop
            The data to be stratified.

        Returns
        -------
           The squeezed data to be stratified.

        Notes
        -----
        The input DataFrame is guaranteeed to have a single column.
        """
        return pop.squeeze(axis=1)


def get_mapped_col_name(col_name: str) -> str:
    """Return a new column name to be used for mapped values"""
    return f"{col_name}_{STRATIFICATION_COLUMN_SUFFIX}"


def get_original_col_name(col_name: str) -> str:
    """Return the original column name given a modified mapped column name."""
    return (
        col_name[: -(len(STRATIFICATION_COLUMN_SUFFIX)) - 1]
        if col_name.endswith(f"_{STRATIFICATION_COLUMN_SUFFIX}")
        else col_name
    )



# mypy: ignore-errors
"""
==================================
The Component Configuration Parser
==================================

The :class:`ComponentConfigurationParser` is responsible for taking a list or
hierarchical :class:`LayeredConfigTree <layered_config_tree.main.LayeredConfigTree>` of components
derived from a model specification yaml file and turning it into a list of
instantiated component objects. When a model specification yaml file is loaded,
the components come in as strings. In order for the simulation to be able to
use these components, they have to be converted into the actual objects they
represent. This occurs via the :meth:`get_components
<ComponentConfigurationParser.get_components>` method of the parser, which is
used anytime a simulation is initialized from a model specification file.

There are three steps to this process.

1. Parsing the model specification's components
2. Validating the arguments and prepping each component
3. Importing and instantiating the actual components

"""

from typing import Dict, List, Tuple, Union

from layered_config_tree.main import LayeredConfigTree

from vivarium.framework.utilities import import_by_path

from ... import Component
from .manager import ComponentConfigError


class ParsingError(ComponentConfigError):
    """Error raised when component configurations are not specified correctly."""

    pass


class ComponentConfigurationParser:
    """Parses component configuration from model specification and initializes
    components.

    To define your own set of parsing rules, you should write a parser class
    that inherits from this class and overrides ``parse_component_config``.
    You can then define a set of parsing rules that turn component configuration
    information into a string which is the full python import path to the class
    followed by a set of parentheses containing initialization arguments.

    For example, say your component configuration specifies ``FancyClass`` from
    the ``important_module`` of your ``made_up_package`` and that ``FancyClass``
    has two initialization parameters, ``'important_thing1'`` and
    ``'important_thing2'``. Your implementation of ``parse_component_config``
    needs to generate the string ``'made_up_package.important_module.FancyClass
    ("important_thing1", "important_thing2")'`` and include it in the list of
    components to generate.

    All classes that are initialized from the ``yaml`` configuration must
    either take no arguments or take arguments specified as strings.

    """

    def get_components(
        self, component_config: Union[LayeredConfigTree, List[str]]
    ) -> List[Component]:
        """Extracts component specifications from configuration information and
        returns initialized components.

        This method encapsulates the three steps described above of parsing,
        validating/prepping, and importing/instantiating.

        The first step of parsing is only done for component configurations that
        come in as a :class:`LayeredConfigTree <layered_config_tree.main.LayeredConfigTree>`.
        Configurations that are provided in the form of a list are already
        assumed to be in the correct form.

        Parameters
        ----------
        component_config
            A hierarchical component specification blob. This configuration
            information needs to be parsable into a full import path and a set
            of initialization arguments by the ``parse_component_config``
            method.

        Returns
        -------
            A list of initialized components.
        """
        if isinstance(component_config, LayeredConfigTree):
            component_list = self.parse_component_config(component_config)
        else:  # Components were specified in a list rather than a tree.
            component_list = [
                self.create_component_from_string(component) for component in component_config
            ]
        return component_list

    def parse_component_config(self, component_config: LayeredConfigTree) -> List[Component]:
        """
        Helper function for parsing a ``LayeredConfigTree`` into a flat list of Components.

        This function converts the ``LayeredConfigTree`` into a dictionary and passes it
        along with an empty prefix list to
        :meth:`process_level <ComponentConfigurationParser.process_level>`. The
        result is a flat list of components.

        Parameters
        ----------
        component_config
            A ``LayeredConfigTree`` representing a hierarchical component specification blob.

        Returns
        -------
            A flat list of Components
        """
        if not component_config:
            return []

        return self.process_level(component_config.to_dict(), [])

    def process_level(
        self, level: Union[str, List[str], Dict[str, Union[Dict, List]]], prefix: List[str]
    ) -> List[Component]:
        """Helper function for parsing hierarchical component configuration into a
        flat list of Components.

        This function recursively walks the hierarchical component configuration,
        treating it like a prefix tree and building the import path prefix. When it
        hits a string, it prepends the built prefix onto the string. If the
        dictionary contains multiple lists, the prefix-prepended lists are
        combined. For example, a component configuration dictionary like the
        following:

        .. code-block:: python

            component_config = {
                'vivarium_examples': {
                    'disease_model': {
                        'population': ['BasePopulation()', 'Mortality()']
                        'disease': ['SIS_DiseaseModel("diarrhea")]
                    }
                }
            }

        would be parsed by this function into the following list:

        .. code-block:: python

            ['vivarium_examples.disease_model.population.BasePopulation()',
             'vivarium_examples.disease_model.population.Mortality()',
             'vivarium_examples.disease_model.disease.SIS_DiseaseModel("diarrhea")']


        Parameters
        ----------
        level
            A level of the hierarchical component specification blob.
        prefix
            A list of strings representing the import path prefix.

        Returns
        -------
            A flat list of Components
        """
        if not level:
            raise ParsingError(
                f"Check your configuration. Component {prefix} should not "
                "be left empty with the header"
            )

        component_list = []
        if isinstance(level, dict):
            for name, child in level.items():
                components = self.process_level(child, prefix + [name])
                component_list.extend(components)
        elif isinstance(level, list):
            for child in level:
                component = self.process_level(child, prefix)
                component_list.extend(component)
        elif isinstance(level, str):
            component = self.create_component_from_string(".".join(prefix + [level]))
            component_list.append(component)
        else:
            raise ParsingError(
                f"Check your configuration. Component {''.join(prefix)}::{level}"
                " should be a string, list, or dictionary."
            )

        return component_list

    def create_component_from_string(self, component_string: str) -> Component:
        """Helper function for creating a component from a string.

        This function takes a string representing a component and turns it into
        an instantiated component object.

        Parameters
        ----------
        component_string
            A string representing the full python import path to the component,
            the component name, and any arguments.

        Returns
        -------
            An instantiated component object.
        """
        component_path, args = self.prep_component(component_string)
        component = self.import_and_instantiate_component(component_path, args)
        return component

    def prep_component(self, component_string: str) -> Tuple[str, Tuple]:
        """Transform component description string into a tuple of component paths
        and required arguments.

        Parameters
        ----------
        component_string
            The component description to transform.

        Returns
        -------
            Component/argument tuple.
        """
        path, args_plus = component_string.split("(")
        cleaned_args = self._clean_args(args_plus[:-1].split(","), path)
        return path, cleaned_args

    @staticmethod
    def _clean_args(args: List, path: str) -> Tuple:
        """Transform component arguments into a tuple, validating that each argument
        is a string.

        Parameters
        ----------
        args
            List of arguments to the component specified at ``path``.
        path
            Path representing the component for which arguments are being cleaned.

        Returns
        -------
            A tuple of arguments, each of which is guaranteed to be a string.
        """
        out = []
        for a in args:
            a = a.strip()
            if not a:
                continue

            not_a_valid_string = len(a) < 3 or not (
                (a[0] == a[-1] == "'") or (a[0] == a[-1] == '"')
            )
            if not_a_valid_string:
                raise ParsingError(f"Invalid component argument {a} for component {path}")

            out.append(a[1:-1])
        return tuple(out)

    @staticmethod
    def import_and_instantiate_component(component_path: str, args: Tuple[str]) -> Component:
        """Transform a tuple representing a Component into an actual instantiated
        component object.

        Parameters
        ----------
        component_path
            A string with the full import path to the component and the
            component name
        args
            A tuple of arguments to the component specified at ``component_path``.

        Returns
        -------
            An instantiated component object.
        """
        return import_by_path(component_path)(*args)



# mypy: ignore-errors
from typing import Any, Dict, List

import pandas as pd

from vivarium import Component
from vivarium.framework.engine import Builder


class Risk(Component):
    CONFIGURATION_DEFAULTS = {
        "risk": {
            "proportion_exposed": 0.3,
        },
    }

    ##############
    # Properties #
    ##############

    @property
    def configuration_defaults(self) -> Dict[str, Any]:
        return {self.risk: self.CONFIGURATION_DEFAULTS["risk"]}

    @property
    def columns_created(self) -> List[str]:
        return [f"{self.risk}_propensity"]

    @property
    def initialization_requirements(self) -> Dict[str, List[str]]:
        return {
            "requires_columns": [],
            "requires_values": [],
            "requires_streams": [self.risk],
        }

    #####################
    # Lifecycle methods #
    #####################

    def __init__(self, risk: str):
        super().__init__()
        self.risk = risk

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder) -> None:
        proportion_exposed = builder.configuration[self.risk].proportion_exposed
        self.base_exposure_threshold = builder.value.register_value_producer(
            f"{self.risk}.base_proportion_exposed",
            source=lambda index: pd.Series(proportion_exposed, index=index),
        )
        self.exposure_threshold = builder.value.register_value_producer(
            f"{self.risk}.proportion_exposed", source=self.base_exposure_threshold
        )

        self.exposure = builder.value.register_value_producer(
            f"{self.risk}.exposure", source=self._exposure
        )
        self.randomness = builder.randomness.get_stream(self.risk)

    ########################
    # Event-driven methods #
    ########################

    def on_initialize_simulants(self, pop_data):
        draw = self.randomness.get_draw(pop_data.index)
        self.population_view.update(pd.Series(draw, name=f"{self.risk}_propensity"))

    ##################################
    # Pipeline sources and modifiers #
    ##################################

    def _exposure(self, index):
        propensity = self.population_view.get(index)[f"{self.risk}_propensity"]
        return self.exposure_threshold(index) > propensity


class RiskEffect(Component):
    CONFIGURATION_DEFAULTS = {
        "risk_effect": {
            "relative_risk": 2,
        },
    }

    ##############
    # Properties #
    ##############

    @property
    def configuration_defaults(self) -> Dict[str, Any]:
        return {self.risk: self.CONFIGURATION_DEFAULTS["risk_effect"]}

    #####################
    # Lifecycle methods #
    #####################

    def __init__(self, risk_name: str, disease_rate: str):
        super().__init__()
        self.risk_name = risk_name
        self.disease_rate = disease_rate
        self.risk = f"effect_of_{risk_name}_on_{disease_rate}"

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder) -> None:
        relative_risk = builder.configuration[self.risk].relative_risk
        self.relative_risk = builder.value.register_value_producer(
            f"{self.risk}.relative_risk",
            source=lambda index: pd.Series(relative_risk, index=index),
        )

        builder.value.register_value_modifier(
            f"{self.disease_rate}.population_attributable_fraction",
            self.population_attributable_fraction,
        )
        builder.value.register_value_modifier(f"{self.disease_rate}", self.rate_adjustment)
        self.base_risk_exposure = builder.value.get_value(
            f"{self.risk_name}.base_proportion_exposed"
        )
        self.actual_risk_exposure = builder.value.get_value(f"{self.risk_name}.exposure")

    ##################################
    # Pipeline sources and modifiers #
    ##################################

    def population_attributable_fraction(self, index):
        exposure = self.base_risk_exposure(index)
        relative_risk = self.relative_risk(index)
        return exposure * (relative_risk - 1) / (exposure * (relative_risk - 1) + 1)

    def rate_adjustment(self, index, rates):
        exposed = self.actual_risk_exposure(index)
        rr = self.relative_risk(index)
        rates[exposed] *= rr[exposed]
        return rates



"""
=======
Manager
=======

A base Manager class to be used to create manager for use in ``vivarium``
simulations.

"""

from typing import TYPE_CHECKING, Any, Dict

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder


class Manager:
    CONFIGURATION_DEFAULTS: Dict[str, Any] = {}
    """A dictionary containing the defaults for any configurations managed by this
    manager. An empty dictionary indicates no managed configurations.

    """

    ##############
    # Properties #
    ##############

    @property
    def configuration_defaults(self) -> Dict[str, Any]:
        """Provides a dictionary containing the defaults for any configurations
        managed by this manager.

        These default values will be stored at the `component_configs` layer of the
        simulation's LayeredConfigTree.

        Returns
        -------
            A dictionary containing the defaults for any configurations managed by
            this manager.
        """
        return self.CONFIGURATION_DEFAULTS

    #####################
    # Lifecycle methods #
    #####################

    def setup(self, builder: "Builder") -> None:
        pass



from vivarium.examples.boids.forces import Alignment, Cohesion, Separation
from vivarium.examples.boids.movement import Movement
from vivarium.examples.boids.neighbors import Neighbors
from vivarium.examples.boids.population import Population
from vivarium.examples.boids.visualization import plot_boids, plot_boids_animated



# mypy: ignore-errors
"""
=============
State Machine
=============

A state machine implementation for use in ``vivarium`` simulations.

"""

from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Tuple

import numpy as np
import pandas as pd

from vivarium import Component

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder
    from vivarium.framework.population import PopulationView
    from vivarium.types import ClockTime


def _next_state(
    index: pd.Index,
    event_time: "ClockTime",
    transition_set: "TransitionSet",
    population_view: "PopulationView",
) -> None:
    """Moves a population between different states using information from a `TransitionSet`.

    Parameters
    ----------
    index
        An iterable of integer labels for the simulants.
    event_time
        When this transition is occurring.
    transition_set
        A set of potential transitions available to the simulants.
    population_view
        A view of the internal state of the simulation.
    """
    if len(transition_set) == 0 or index.empty:
        return

    outputs, decisions = transition_set.choose_new_state(index)
    groups = _groupby_new_state(index, outputs, decisions)

    if groups:
        for output, affected_index in sorted(groups, key=lambda x: str(x[0])):
            if output == "null_transition":
                pass
            elif isinstance(output, Transient):
                if not isinstance(output, State):
                    raise ValueError("Invalid transition output: {}".format(output))
                output.transition_effect(affected_index, event_time, population_view)
                output.next_state(affected_index, event_time, population_view)
            elif isinstance(output, State):
                output.transition_effect(affected_index, event_time, population_view)
            else:
                raise ValueError("Invalid transition output: {}".format(output))


def _groupby_new_state(
    index: pd.Index, outputs: List, decisions: pd.Series
) -> List[Tuple[str, pd.Index]]:
    """Groups the simulants in the index by their new output state.

    Parameters
    ----------
    index
        An iterable of integer labels for the simulants.
    outputs
        A list of possible output states.
    decisions
        A series containing the name of the next state for each simulant in the
        index.

    Returns
    -------
        The first item in each tuple is the name of an output state and the
        second item is a `pandas.Index` representing the simulants to transition
        into that state.
    """
    groups = pd.Series(index).groupby(
        pd.Categorical(decisions.values, categories=outputs), observed=False
    )
    return [(output, pd.Index(sub_group.values)) for output, sub_group in groups]


class Trigger(Enum):
    NOT_TRIGGERED = 0
    START_INACTIVE = 1
    START_ACTIVE = 2


def _process_trigger(trigger):
    if trigger == Trigger.NOT_TRIGGERED:
        return None, False
    elif trigger == Trigger.START_INACTIVE:
        return pd.Index([]), False
    elif trigger == Trigger.START_ACTIVE:
        return pd.Index([]), True
    else:
        raise ValueError("Invalid trigger state provided: {}".format(trigger))


class Transition(Component):
    """A process by which an entity might change into a particular state.

    Attributes
    ----------
    input_state
        The start state of the entity that undergoes the transition.
    output_state
        The end state of the entity that undergoes the transition.
    probability_func
        A method or function that describing the probability of this
        transition occurring.
    triggered
        A flag indicating whether this transition is triggered by some event.
    """

    #####################
    # Lifecycle methods #
    #####################

    def __init__(
        self,
        input_state: "State",
        output_state: "State",
        probability_func: Callable[[pd.Index], pd.Series] = lambda index: pd.Series(
            1.0, index=index
        ),
        triggered=Trigger.NOT_TRIGGERED,
    ):
        super().__init__()
        self.input_state = input_state
        self.output_state = output_state
        self._probability = probability_func
        self._active_index, self.start_active = _process_trigger(triggered)

    ##################
    # Public methods #
    ##################

    def set_active(self, index: pd.Index) -> None:
        if self._active_index is None:
            raise ValueError(
                "This transition is not triggered.  An active index cannot be set or modified."
            )
        else:
            self._active_index = self._active_index.union(pd.Index(index))

    def set_inactive(self, index: pd.Index) -> None:
        if self._active_index is None:
            raise ValueError(
                "This transition is not triggered.  An active index cannot be set or modified."
            )
        else:
            self._active_index = self._active_index.difference(pd.Index(index))

    def probability(self, index: pd.Index) -> pd.Series:
        if self._active_index is None:
            return self._probability(index)

        index = pd.Index(index)
        activated_index = self._active_index.intersection(index)
        null_index = index.difference(self._active_index)
        activated = pd.Series(self._probability(activated_index), index=activated_index)
        null = pd.Series(np.zeros(len(null_index), dtype=float), index=null_index)
        return activated.append(null)


class State(Component):
    """An abstract representation of a particular position in a state space.

    Attributes
    ----------
    state_id
        The name of this state. This should be unique
    transition_set
        A container for potential transitions out of this state.

    """

    ##############
    # Properties #
    ##############

    @property
    def model(self) -> str:
        return self._model

    #####################
    # Lifecycle methods #
    #####################

    def __init__(self, state_id: str, allow_self_transition: bool = False):
        super().__init__()
        self.state_id = state_id
        self.transition_set = TransitionSet(
            self.state_id, allow_self_transition=allow_self_transition
        )
        self._model = None
        self._sub_components = [self.transition_set]

    ##################
    # Public methods #
    ##################

    def set_model(self, model_name: str) -> None:
        """Defines the column name for the model this state belongs to"""
        self._model = model_name

    def next_state(
        self, index: pd.Index, event_time: "ClockTime", population_view: "PopulationView"
    ) -> None:
        """Moves a population between different states.

        Parameters
        ----------
        index
            An iterable of integer labels for the simulants.
        event_time
            When this transition is occurring.
        population_view
            A view of the internal state of the simulation.
        """
        return _next_state(index, event_time, self.transition_set, population_view)

    def transition_effect(
        self, index: pd.Index, event_time: "ClockTime", population_view: "PopulationView"
    ) -> None:
        """Updates the simulation state and triggers any side-effects associated with entering this state.

        Parameters
        ----------
        index
            An iterable of integer labels for the simulants.
        event_time
            The time at which this transition occurs.
        population_view
            A view of the internal state of the simulation.
        """
        population_view.update(pd.Series(self.state_id, index=index))
        self.transition_side_effect(index, event_time)

    def cleanup_effect(self, index: pd.Index, event_time: "ClockTime") -> None:
        pass

    def add_transition(self, transition: Transition) -> None:
        """Adds a transition to this state and its `TransitionSet`.

        Parameters
        ----------
        transition
            The transition to add
        """
        self.transition_set.append(transition)

    def allow_self_transitions(self) -> None:
        self.transition_set.allow_null_transition = True

    ##################
    # Helper methods #
    ##################

    def transition_side_effect(self, index: pd.Index, event_time: "ClockTime") -> None:
        pass


class Transient:
    """Used to tell _next_state to transition a second time."""

    pass


class TransientState(State, Transient):
    pass


class TransitionSet(Component):
    """A container for state machine transitions.

    Attributes
    ----------
    state_id
        The unique name of the state that instantiated this TransitionSet. Typically
        a string but any object implementing __str__ will do.
    allow_null_transition
        Specified whether it is possible not to transition on a given time-step
    transitions
        A list of transitions that can be taken from this state.
    random
        The randomness stream.

    """

    ##############
    # Properties #
    ##############

    @property
    def name(self) -> str:
        return f"transition_set.{self.state_id}"

    #####################
    # Lifecycle methods #
    #####################

    def __init__(
        self, state_id: str, *transitions: Transition, allow_self_transition: bool = False
    ):
        super().__init__()
        self.state_id = state_id
        self.allow_null_transition = allow_self_transition
        self.transitions = []
        self._sub_components = self.transitions

        self.extend(transitions)

    def setup(self, builder: "Builder") -> None:
        """Performs this component's simulation setup and return sub-components.

        Parameters
        ----------
        builder
            Interface to several simulation tools including access to common random
            number generation, in particular.
        """
        self.random = builder.randomness.get_stream(self.name)

    ##################
    # Public methods #
    ##################

    def choose_new_state(self, index: pd.Index) -> Tuple[List, pd.Series]:
        """Chooses a new state for each simulant in the index.

        Parameters
        ----------
        index
            An iterable of integer labels for the simulants.

        Returns
        -------
            A tuple of the possible end states of this set of transitions and a
            series containing the name of the next state for each simulant
            in the index.
        """
        outputs, probabilities = zip(
            *[
                (transition.output_state, np.array(transition.probability(index)))
                for transition in self.transitions
            ]
        )
        probabilities = np.transpose(probabilities)
        outputs, probabilities = self._normalize_probabilities(outputs, probabilities)
        return outputs, self.random.choice(index, outputs, probabilities)

    def append(self, transition: Transition) -> None:
        if not isinstance(transition, Transition):
            raise TypeError(
                "TransitionSet must contain only Transition objects. "
                f"Check constructor arguments: {self}"
            )
        self.transitions.append(transition)

    def extend(self, transitions: Iterable[Transition]) -> None:
        for transition in transitions:
            self.append(transition)

    ##################
    # Helper methods #
    ##################

    def _normalize_probabilities(self, outputs, probabilities):
        """Normalize probabilities to sum to 1 and add a null transition.

        Parameters
        ----------
        outputs
            List of possible end states corresponding to this containers
            transitions.
        probabilities
            A set of probability weights whose columns correspond to the end
            states in `outputs` and whose rows correspond to each simulant
            undergoing the transition.

        Returns
        -------
            A tuple of the original output list expanded to include a null transition
            (a transition back to the starting state) if requested and the original
            probabilities rescaled to sum to 1 and potentially expanded to include
            a null transition weight.
        """
        outputs = list(outputs)

        # This is mainly for flexibility with the triggered transitions.
        # We may have multiple out transitions from a state where one of them
        # is gated until some criteria is met.  After the criteria is
        # met, the gated transition becomes the default (likely as opposed
        # to a self transition).
        default_transition_count = np.sum(probabilities == 1, axis=1)
        if np.any(default_transition_count > 1):
            raise ValueError("Multiple transitions specified with probability 1.")
        has_default = default_transition_count == 1
        total = np.sum(probabilities, axis=1)
        probabilities[has_default] /= total[has_default, np.newaxis]

        total = np.sum(probabilities, axis=1)  # All totals should be ~<= 1 at this point.
        if self.allow_null_transition:
            if np.any(total > 1 + 1e-08):  # Accommodate rounding errors
                raise ValueError(
                    f"Null transition requested with un-normalized "
                    f"probability weights: {probabilities}"
                )
            total[total > 1] = 1  # Correct allowed rounding errors.
            probabilities = np.concatenate(
                [probabilities, (1 - total)[:, np.newaxis]], axis=1
            )
            outputs.append("null_transition")
        else:
            if np.any(total == 0):
                raise ValueError("No valid transitions for some simulants.")
            else:  # total might be less than zero in some places
                probabilities /= total[:, np.newaxis]

        return outputs, probabilities

    def __iter__(self):
        return iter(self.transitions)

    def __len__(self):
        return len(self.transitions)

    def __hash__(self):
        return hash(id(self))


class Machine(Component):
    """A collection of states and transitions between those states.

    Attributes
    ----------
    states
        The collection of states represented by this state machine.
    state_column
        A label for the piece of simulation state governed by this state machine.

    """

    ##############
    # Properties #
    ##############

    @property
    def sub_components(self):
        return self.states

    @property
    def columns_required(self) -> Optional[List[str]]:
        return [self.state_column]

    #####################
    # Lifecycle methods #
    #####################

    def __init__(self, state_column: str, states: Iterable[State] = ()):
        super().__init__()
        self.states = []
        self.state_column = state_column
        if states:
            self.add_states(states)

    ##################
    # Public methods #
    ##################

    def add_states(self, states: Iterable[State]) -> None:
        for state in states:
            self.states.append(state)
            state.set_model(self.state_column)

    def transition(self, index: pd.Index, event_time: "ClockTime") -> None:
        """Finds the population in each state and moves them to the next state.

        Parameters
        ----------
        index
            An iterable of integer labels for the simulants.
        event_time
            The time at which this transition occurs.
        """
        for state, affected in self._get_state_pops(index):
            if not affected.empty:
                state.next_state(
                    affected.index,
                    event_time,
                    self.population_view.subview([self.state_column]),
                )

    def cleanup(self, index: pd.Index, event_time: "ClockTime") -> None:
        for state, affected in self._get_state_pops(index):
            if not affected.empty:
                state.cleanup_effect(affected.index, event_time)

    def _get_state_pops(self, index: pd.Index) -> List[Tuple[State, pd.DataFrame]]:
        population = self.population_view.get(index)
        return [
            (state, population[population[self.state_column] == state.state_id])
            for state in self.states
        ]

    ##################
    # Helper methods #
    ##################

    def get_initialization_parameters(self) -> Dict[str, Any]:
        """
        Gets the values of the state column specified in the __init__`.

        Returns
        -------
            The value of the state column.

        Notes
        -----
        This retrieves the value of the attribute at the time of calling
        which is not guaranteed to be the same as the original value.
        """

        return {"state_column": self.state_column}



from vivarium.framework.results.interface import ResultsInterface
from vivarium.framework.results.manager import ResultsManager
from vivarium.framework.results.observation import VALUE_COLUMN
from vivarium.framework.results.observer import Observer



# mypy: ignore-errors
"""
============================
The Component Manager System
============================

The :mod:`vivarium` component manager system is responsible for maintaining a
reference to all of the managers and components in a simulation, providing an
interface for adding additional components or managers, and applying default
configurations and initiating the ``setup`` stage of the lifecycle. This module
provides the default implementation and interface.

The :class:`ComponentManager` is the first plugin loaded by the
:class:`SimulationContext <vivarium.framework.engine.SimulationContext>`
and managers and components are given to it by the context. It is called on to
setup everything it holds when the context itself is setup.

"""

import inspect
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Sequence, Tuple, Union

from layered_config_tree import (
    ConfigurationError,
    DuplicatedConfigurationError,
    LayeredConfigTree,
)

from vivarium import Component
from vivarium.exceptions import VivariumError
from vivarium.framework.lifecycle import LifeCycleManager
from vivarium.manager import Manager

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder


class ComponentConfigError(VivariumError):
    """Error while interpreting configuration file or initializing components"""

    pass


class OrderedComponentSet:
    """A container for Vivarium components.

    It preserves ordering, enforces uniqueness by name, and provides a
    subset of set-like semantics.

    """

    def __init__(self, *args: Union[Component, Manager]):
        self.components: List[Union[Component, Manager]] = []
        if args:
            self.update(args)

    def add(self, component: Component) -> None:
        if component in self:
            raise ComponentConfigError(
                f"Attempting to add a component with duplicate name: {component}"
            )
        self.components.append(component)

    def update(
        self,
        components: Union[List[Union[Component, Manager]], Tuple[Union[Component, Manager]]],
    ) -> None:
        for c in components:
            self.add(c)

    def pop(self) -> Union[Component, Manager]:
        component = self.components.pop(0)
        return component

    def __contains__(self, component: Union[Component, Manager]) -> bool:
        if not hasattr(component, "name"):
            raise ComponentConfigError(f"Component {component} has no name attribute")
        return component.name in [c.name for c in self.components]

    def __iter__(self) -> Iterator[Union[Component, Manager]]:
        return iter(self.components)

    def __len__(self) -> int:
        return len(self.components)

    def __bool__(self) -> bool:
        return bool(self.components)

    def __add__(self, other: "OrderedComponentSet") -> "OrderedComponentSet":
        return OrderedComponentSet(*(self.components + other.components))

    def __eq__(self, other: "OrderedComponentSet") -> bool:
        try:
            return type(self) is type(other) and [c.name for c in self.components] == [
                c.name for c in other.components
            ]
        except TypeError:
            return False

    def __getitem__(self, index: int) -> Any:
        return self.components[index]

    def __repr__(self):
        return f"OrderedComponentSet({[c.name for c in self.components]})"


class ComponentManager(Manager):
    """Manages the initialization and setup of :mod:`vivarium` components.

    Maintains references to all components and managers in a :mod:`vivarium`
    simulation, applies their default configuration and initiates their
    ``setup`` life-cycle stage.

    The component manager maintains a separate list of managers and components
    and provides methods for adding to these lists and getting members that
    correspond to a specific type.  It also initiates the ``setup`` lifecycle
    phase for all components and managers it controls. This is done first for
    managers and then components, and involves applying default configurations
    and calling the object's ``setup`` method.

    """

    def __init__(self):
        self._managers = OrderedComponentSet()
        self._components = OrderedComponentSet()
        self.configuration = None
        self.lifecycle = None

    @property
    def name(self):
        """The name of this component."""
        return "component_manager"

    def setup(
        self, configuration: LayeredConfigTree, lifecycle_manager: LifeCycleManager
    ) -> None:
        """Called by the simulation context."""
        self.configuration = configuration
        self.lifecycle = lifecycle_manager

        self.lifecycle.add_constraint(
            self.get_components_by_type,
            restrict_during=["initialization", "population_creation"],
        )
        self.lifecycle.add_constraint(
            self.get_component, restrict_during=["population_creation"]
        )
        self.lifecycle.add_constraint(
            self.list_components, restrict_during=["initialization"]
        )

    def add_managers(self, managers: Union[List[Manager], Tuple[Manager]]) -> None:
        """Registers new managers with the component manager.

        Managers are configured and setup before components.

        Parameters
        ----------
        managers
            Instantiated managers to register.
        """
        for m in self._flatten(list(managers)):
            self.apply_configuration_defaults(m)
            self._managers.add(m)

    def add_components(self, components: Union[List[Component], Tuple[Component]]) -> None:
        """Register new components with the component manager.

        Components are configured and setup after managers.

        Parameters
        ----------
        components
            Instantiated components to register.
        """
        for c in self._flatten(list(components)):
            self.apply_configuration_defaults(c)
            self._components.add(c)

    def get_components_by_type(
        self, component_type: Union[type, Sequence[type]]
    ) -> List[Component]:
        """Get all components that are an instance of ``component_type``.

        Parameters
        ----------
        component_type
            A component type.

        Returns
        -------
            A list of components of type ``component_type``.
        """
        # Convert component_type to a tuple for isinstance
        return [c for c in self._components if isinstance(c, tuple(component_type))]

    def get_component(self, name: str) -> Component:
        """Get the component with name ``name``.

        Names are guaranteed to be unique.

        Parameters
        ----------
        name
            A component name.

        Returns
        -------
            A component that has name ``name``.

        Raises
        ------
        ValueError
            No component exists in the component manager with ``name``.
        """
        for c in self._components:
            if c.name == name:
                return c
        raise ValueError(f"No component found with name {name}")

    def list_components(self) -> Dict[str, Component]:
        """Get a mapping of component names to components held by the manager.

        Returns
        -------
            A mapping of component names to components.

        """
        return {c.name: c for c in self._components}

    def setup_components(self, builder: "Builder") -> None:
        """Separately configure and set up the managers and components held by
        the component manager, in that order.

        The setup process involves applying default configurations and then
        calling the manager or component's setup method. This can result in new
        components as a side effect of setup because components themselves have
        access to this interface through the builder in their setup method.

        Parameters
        ----------
        builder
            Interface to several simulation tools.
        """
        self._setup_components(builder, self._managers + self._components)

    def apply_configuration_defaults(self, component: Union[Component, Manager]) -> None:
        try:
            self.configuration.update(
                component.configuration_defaults,
                layer="component_configs",
                source=component.name,
            )
        except DuplicatedConfigurationError as e:
            new_name, new_file = component.name, self._get_file(component)
            old_name, old_file = e.source, self._get_file(self.get_component(e.source))

            raise ComponentConfigError(
                f"Component {new_name} in file {new_file} is attempting to "
                f"set the configuration value {e.value_name}, but it has already "
                f"been set by {old_name} in file {old_file}."
            )
        except ConfigurationError as e:
            new_name, new_file = component.name, self._get_file(component)
            raise ComponentConfigError(
                f"Component {new_name} in file {new_file} is attempting to "
                f"alter the structure of the configuration at key {e.value_name}. "
                f"This happens if one component attempts to set a value at an interior "
                f"configuration key or if it attempts to turn an interior key into a "
                f"configuration value."
            )

    @staticmethod
    def _get_file(component: Union[Component, Manager]) -> str:
        if component.__module__ == "__main__":
            # This is defined directly in a script or notebook so there's no
            # file to attribute it to.
            return "__main__"
        else:
            return inspect.getfile(component.__class__)

    @staticmethod
    def _flatten(
        components: List[Union[Component, Manager]]
    ) -> List[Union[Component, Manager]]:
        out = []
        components = components[::-1]
        while components:
            current = components.pop()
            if isinstance(current, (list, tuple)):
                components.extend(current[::-1])
            elif isinstance(current, Component):
                components.extend(current.sub_components[::-1])
                out.append(current)
            elif isinstance(current, Manager):
                out.append(current)
            else:
                raise TypeError(
                    "Expected Component, Manager, List, or Tuple. "
                    f"Got {type(current)}: {current}"
                )
        return out

    @staticmethod
    def _setup_components(builder: "Builder", components: OrderedComponentSet):
        for component in components:
            if isinstance(component, Component):
                component.setup_component(builder)
            elif isinstance(component, Manager):
                component.setup(builder)

    def __repr__(self):
        return "ComponentManager()"


class ComponentInterface:
    """The builder interface for the component manager system.

    This class defines component manager methods a ``vivarium`` component can
    access from the builder. It provides methods for querying and adding components
    to the :class:`ComponentManager`.
    """

    def __init__(self, manager: ComponentManager):
        self._manager = manager

    def get_component(self, name: str) -> Component:
        """Get the component that has ``name`` if presently held by the component
        manager. Names are guaranteed to be unique.

        Parameters
        ----------
        name
            A component name.

        Returns
        -------
            A component that has name ``name``.
        """
        return self._manager.get_component(name)

    def get_components_by_type(
        self, component_type: Union[type, Sequence[type]]
    ) -> List[Component]:
        """Get all components that are an instance of ``component_type``.

        Parameters
        ----------
        component_type
            A component type to retrieve, compared against internal components
            using isinstance().

        Returns
        -------
            A list of components of type ``component_type``.
        """
        return self._manager.get_components_by_type(component_type)

    def list_components(self) -> Dict[str, Component]:
        """Get a mapping of component names to components held by the manager.

        Returns
        -------
            A dictionary mapping component names to components.
        """
        return self._manager.list_components()
